"""
  The Job Sanity executor screens jobs for the following problems:

     - Output data already exists
     - Problematic JDL
     - Jobs with too much input data e.g. > 100 files
     - Jobs with input data incorrectly specified e.g. castor:/
     - Input sandbox not correctly uploaded.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import re

from DIRAC import S_OK, S_ERROR

from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.WorkloadManagementSystem.Client.SandboxStoreClient import SandboxStoreClient
from DIRAC.WorkloadManagementSystem.Executor.Base.OptimizerExecutor import OptimizerExecutor
from DIRAC.WorkloadManagementSystem.Client import JobMinorStatus


class JobSanity(OptimizerExecutor):
    """
    The specific Optimizer must provide the following methods:
      - optimizeJob() - the main method called for each job
    and it can provide:
      - initializeOptimizer() before each execution cycle
    """

    @classmethod
    def initializeOptimizer(cls):
        """Initialize specific parameters for JobSanityAgent."""
        cls.sandboxClient = SandboxStoreClient(useCertificates=True, smdb=True)
        return S_OK()

    def optimizeJob(self, jid, jobState):
        """This method controls the order and presence of
        each sanity check for submitted jobs. This should
        be easily extended in the future to accommodate
        any other potential checks.
        """
        # Job JDL check
        result = jobState.getAttribute("JobType")
        if not result["OK"]:
            self.jobLog.error("Failed to get job type attribute", result["Message"])
            return result
        jobType = result["Value"].lower()

        result = jobState.getManifest()
        if not result["OK"]:
            self.jobLog.error("Failed to get job manifest", result["Message"])
            return result
        manifest = result["Value"]

        # Input data check
        if self.ex_getOption("InputDataCheck", True):
            voName = manifest.getOption("VirtualOrganization", "")
            if not voName:
                return S_ERROR("No VirtualOrganization defined in manifest")
            result = self.checkInputData(jobState, jobType, voName)
            if not result["OK"]:
                self.jobLog.error("Failed to check input data", result["Message"])
                return result
            self.jobLog.info("Found LFNs", result["Value"])

        # Input Sandbox check
        if self.ex_getOption("InputSandboxCheck", True):
            result = self.checkInputSandbox(jobState, manifest)
            if not result["OK"]:
                self.jobLog.error("Failed to check input sandbox", result["Message"])
                return result
            self.jobLog.info("Assigned ISBs", result["Value"])

        return self.setNextOptimizer(jobState)

    def checkInputData(self, jobState, jobType, voName):
        """This method checks both the amount of input
        datasets for the job and whether the LFN conventions
        are correct.
        """

        result = jobState.getInputData()
        if not result["OK"]:
            self.jobLog.warn("Failed to get input data from JobDB", result["Message"])
            return S_ERROR("Input Data Specification")

        data = result["Value"]  # seems to be [''] when null, which isn't an empty list ;)
        data = [lfn.strip() for lfn in data if lfn.strip()]
        if not data:
            return S_OK(0)

        self.jobLog.debug("Input data requirement will be checked")
        self.jobLog.debug("Data is:\n\t%s" % "\n\t".join(data))

        voRE = re.compile("^(LFN:)?/%s/" % voName)

        for lfn in data:
            if not voRE.match(lfn):
                return S_ERROR(JobMinorStatus.INPUT_INCORRECT)
            if lfn.find("//") > -1:
                return S_ERROR(JobMinorStatus.INPUT_CONTAINS_SLASHES)

        # only check limit for user jobs
        if jobType == "user":
            maxLFNs = self.ex_getOption("MaxInputDataPerJob", 100)
            if len(data) > maxLFNs:
                return S_ERROR("Exceeded Maximum Dataset Limit (%s)" % maxLFNs)
        return S_OK(len(data))

    def checkInputSandbox(self, jobState, manifest):
        """The number of input sandbox files, as specified in the job
        JDL are checked in the JobDB.
        """
        result = jobState.getAttributes(["Owner", "OwnerDN", "OwnerGroup", "DIRACSetup"])
        if not result["OK"]:
            self.jobLog.error("Failed to get job attributes", result["Message"])
            return result
        attDict = result["Value"]
        ownerName = attDict["Owner"]
        if not ownerName:
            ownerDN = attDict["OwnerDN"]
            if not ownerDN:
                return S_ERROR("Missing OwnerDN")
            result = Registry.getUsernameForDN(ownerDN)
            if not result["OK"]:
                self.jobLog.error("Failed to get user name from DN", result["Message"])
                return result
            ownerName = result["Value"]
        ownerGroup = attDict["OwnerGroup"]
        if not ownerGroup:
            return S_ERROR("Missing OwnerGroup")
        jobSetup = attDict["DIRACSetup"]
        if not jobSetup:
            return S_ERROR("Missing DIRACSetup")

        isbList = manifest.getOption("InputSandbox", [])
        sbsToAssign = []
        for isb in isbList:
            if isb.startswith("SB:"):
                self.jobLog.debug("Found a sandbox", isb)
                sbsToAssign.append((isb, "Input"))
            if isb.startswith("LFN:"):
                self.jobLog.debug("Found a LFN sandbox", isb)
                if isb[4] != "/":  # the LFN does not start with /
                    return S_ERROR("LFNs should always start with '/'")
        numSBsToAssign = len(sbsToAssign)
        if not numSBsToAssign:
            return S_OK(0)
        self.jobLog.info("Assigning sandboxes", "(%s on behalf of %s@%s)" % (numSBsToAssign, ownerName, ownerGroup))
        result = self.sandboxClient.assignSandboxesToJob(jobState.jid, sbsToAssign, ownerName, ownerGroup, jobSetup)
        if not result["OK"]:
            self.jobLog.error("Could not assign sandboxes in the SandboxStore")
            return S_ERROR("Cannot assign sandbox to job")
        assigned = result["Value"]
        if assigned != numSBsToAssign:
            self.jobLog.error("Could not assign all sandboxes", "(%s). Only assigned %s" % (numSBsToAssign, assigned))
        return S_OK(numSBsToAssign)



"""Class for making requests to a ComponentMonitoring Service."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Base.Client import Client, createClient


@createClient("Framework/ComponentMonitoring")
class ComponentMonitoringClient(Client):
    def __init__(self, **kwargs):
        """
        Constructor function
        """

        super(ComponentMonitoringClient, self).__init__(**kwargs)
        self.setServer("Framework/ComponentMonitoring")



""" DISET request handler base class for the TransformationDB.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Utilities.DEncode import ignoreEncodeWarning
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.TransformationSystem.Client import TransformationFilesStatus
from DIRAC.WorkloadManagementSystem.Client import JobStatus


transTypes = list(six.string_types) + list(six.integer_types)

__RCSID__ = "$Id$"

TASKS_STATE_NAMES = ["TotalCreated", "Created"] + JobStatus.JOB_STATES
FILES_STATE_NAMES = ["PercentProcessed", "Total"] + TransformationFilesStatus.TRANSFORMATION_FILES_STATES


class TransformationManagerHandler(RequestHandler):
    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        """Initialization of DB object"""

        try:
            result = ObjectLoader().loadObject("TransformationSystem.DB.TransformationDB", "TransformationDB")
            if not result["OK"]:
                return result
            cls.transformationDB = result["Value"]()

        except RuntimeError as excp:
            return S_ERROR("Can't connect to TransformationDB: %s" % excp)

        return S_OK()

    types_getCounters = [six.string_types, list, dict]

    @classmethod
    def export_getCounters(cls, table, attrList, condDict, older=None, newer=None, timeStamp=None):
        return cls.transformationDB.getCounters(
            table, attrList, condDict, older=older, newer=newer, timeStamp=timeStamp
        )

    ####################################################################
    #
    # These are the methods to manipulate the transformations table
    #

    types_addTransformation = [
        six.string_types,
        six.string_types,
        six.string_types,
        six.string_types,
        six.string_types,
        six.string_types,
        six.string_types,
    ]

    def export_addTransformation(
        self,
        transName,
        description,
        longDescription,
        transType,
        plugin,
        agentType,
        fileMask,
        transformationGroup="General",
        groupSize=1,
        inheritedFrom=0,
        body="",
        maxTasks=0,
        eventsPerTask=0,
        addFiles=True,
        inputMetaQuery=None,
        outputMetaQuery=None,
    ):
        #    authorDN = self._clientTransport.peerCredentials['DN']
        #    authorGroup = self._clientTransport.peerCredentials['group']
        credDict = self.getRemoteCredentials()
        authorDN = credDict.get("DN", credDict.get("CN"))
        authorGroup = credDict.get("group")
        res = self.transformationDB.addTransformation(
            transName,
            description,
            longDescription,
            authorDN,
            authorGroup,
            transType,
            plugin,
            agentType,
            fileMask,
            transformationGroup=transformationGroup,
            groupSize=groupSize,
            inheritedFrom=inheritedFrom,
            body=body,
            maxTasks=maxTasks,
            eventsPerTask=eventsPerTask,
            addFiles=addFiles,
            inputMetaQuery=inputMetaQuery,
            outputMetaQuery=outputMetaQuery,
        )
        if res["OK"]:
            self.log.info("Added transformation", res["Value"])
        return res

    types_deleteTransformation = [transTypes]

    def export_deleteTransformation(self, transName):
        credDict = self.getRemoteCredentials()
        authorDN = credDict.get("DN", credDict.get("CN"))
        # authorDN = self._clientTransport.peerCredentials['DN']
        return self.transformationDB.deleteTransformation(transName, author=authorDN)

    types_cleanTransformation = [transTypes]

    def export_cleanTransformation(self, transName):
        credDict = self.getRemoteCredentials()
        authorDN = credDict.get("DN", credDict.get("CN"))
        # authorDN = self._clientTransport.peerCredentials['DN']
        return self.transformationDB.cleanTransformation(transName, author=authorDN)

    types_setTransformationParameter = [transTypes, six.string_types]

    def export_setTransformationParameter(self, transName, paramName, paramValue):
        credDict = self.getRemoteCredentials()
        authorDN = credDict.get("DN", credDict.get("CN"))
        # authorDN = self._clientTransport.peerCredentials['DN']
        return self.transformationDB.setTransformationParameter(transName, paramName, paramValue, author=authorDN)

    types_deleteTransformationParameter = [transTypes, six.string_types]

    @classmethod
    def export_deleteTransformationParameter(cls, transName, paramName):
        # credDict = self.getRemoteCredentials()
        # authorDN = credDict[ 'DN' ]
        # authorDN = self._clientTransport.peerCredentials['DN']
        return cls.transformationDB.deleteTransformationParameter(transName, paramName)

    types_getTransformations = []

    @classmethod
    def export_getTransformations(
        cls,
        condDict=None,
        older=None,
        newer=None,
        timeStamp="CreationDate",
        orderAttribute=None,
        limit=None,
        extraParams=False,
        offset=None,
        columns=None,
    ):
        if not condDict:
            condDict = {}
        return cls.transformationDB.getTransformations(
            condDict=condDict,
            older=older,
            newer=newer,
            timeStamp=timeStamp,
            orderAttribute=orderAttribute,
            limit=limit,
            extraParams=extraParams,
            offset=offset,
            columns=columns,
        )

    types_getTransformation = [transTypes]

    @classmethod
    def export_getTransformation(cls, transName, extraParams=False):
        return cls.transformationDB.getTransformation(transName, extraParams=extraParams)

    types_getTransformationParameters = [transTypes, [six.string_types, list]]

    @classmethod
    def export_getTransformationParameters(cls, transName, parameters):
        return cls.transformationDB.getTransformationParameters(transName, parameters)

    types_getTransformationWithStatus = [[six.string_types, list, tuple]]

    @classmethod
    def export_getTransformationWithStatus(cls, status):
        return cls.transformationDB.getTransformationWithStatus(status)

    ####################################################################
    #
    # These are the methods to manipulate the TransformationFiles tables
    #

    types_addFilesToTransformation = [transTypes, [list, tuple]]

    @classmethod
    def export_addFilesToTransformation(cls, transName, lfns):
        return cls.transformationDB.addFilesToTransformation(transName, lfns)

    types_addTaskForTransformation = [transTypes]

    @classmethod
    def export_addTaskForTransformation(cls, transName, lfns=[], se="Unknown"):
        return cls.transformationDB.addTaskForTransformation(transName, lfns=lfns, se=se)

    types_setFileStatusForTransformation = [transTypes, dict]

    @classmethod
    @ignoreEncodeWarning
    def export_setFileStatusForTransformation(cls, transName, dictOfNewFilesStatus):
        """Sets the file status for the transformation.

        The dictOfNewFilesStatus is a dictionary with the form:
        {12345: ('StatusA', errorA), 6789: ('StatusB',errorB),  ... } where the keys are fileIDs
        The tuple may be a string with only the status if the client was from an older version
        """

        if not dictOfNewFilesStatus:
            return S_OK({})

        statusSample = list(dictOfNewFilesStatus.values())[0]
        if isinstance(statusSample, (list, tuple)) and len(statusSample) == 2:
            newStatusForFileIDs = dictOfNewFilesStatus
        else:
            return S_ERROR("Status field should be two values")

        res = cls.transformationDB._getConnectionTransID(False, transName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]

        return cls.transformationDB.setFileStatusForTransformation(transID, newStatusForFileIDs, connection=connection)

    types_getTransformationStats = [transTypes]

    @classmethod
    def export_getTransformationStats(cls, transName):
        return cls.transformationDB.getTransformationStats(transName)

    types_getTransformationFilesCount = [transTypes, six.string_types]

    @classmethod
    def export_getTransformationFilesCount(cls, transName, field, selection={}):
        return cls.transformationDB.getTransformationFilesCount(transName, field, selection=selection)

    types_getTransformationFiles = []

    @classmethod
    def export_getTransformationFiles(
        cls, condDict=None, older=None, newer=None, timeStamp="LastUpdate", orderAttribute=None, limit=None, offset=None
    ):
        if not condDict:
            condDict = {}
        return cls.transformationDB.getTransformationFiles(
            condDict=condDict,
            older=older,
            newer=newer,
            timeStamp=timeStamp,
            orderAttribute=orderAttribute,
            limit=limit,
            offset=offset,
            connection=False,
        )

    ####################################################################
    #
    # These are the methods to manipulate the TransformationTasks table
    #

    types_getTransformationTasks = []

    @classmethod
    def export_getTransformationTasks(
        cls,
        condDict=None,
        older=None,
        newer=None,
        timeStamp="CreationTime",
        orderAttribute=None,
        limit=None,
        inputVector=False,
        offset=None,
    ):
        if not condDict:
            condDict = {}
        return cls.transformationDB.getTransformationTasks(
            condDict=condDict,
            older=older,
            newer=newer,
            timeStamp=timeStamp,
            orderAttribute=orderAttribute,
            limit=limit,
            inputVector=inputVector,
            offset=offset,
        )

    types_setTaskStatus = [transTypes, [list] + list(six.integer_types), six.string_types]

    @classmethod
    def export_setTaskStatus(cls, transName, taskID, status):
        return cls.transformationDB.setTaskStatus(transName, taskID, status)

    types_setTaskStatusAndWmsID = [transTypes, list(six.integer_types), six.string_types, six.string_types]

    @classmethod
    def export_setTaskStatusAndWmsID(cls, transName, taskID, status, taskWmsID):
        return cls.transformationDB.setTaskStatusAndWmsID(transName, taskID, status, taskWmsID)

    types_getTransformationTaskStats = [transTypes]

    @classmethod
    def export_getTransformationTaskStats(cls, transName):
        return cls.transformationDB.getTransformationTaskStats(transName)

    types_deleteTasks = [transTypes, list(six.integer_types), list(six.integer_types)]

    def export_deleteTasks(self, transName, taskMin, taskMax):
        credDict = self.getRemoteCredentials()
        authorDN = credDict.get("DN", credDict.get("CN"))
        # authorDN = self._clientTransport.peerCredentials['DN']
        return self.transformationDB.deleteTasks(transName, taskMin, taskMax, author=authorDN)

    types_extendTransformation = [transTypes, list(six.integer_types)]

    def export_extendTransformation(self, transName, nTasks):
        credDict = self.getRemoteCredentials()
        authorDN = credDict.get("DN", credDict.get("CN"))
        # authorDN = self._clientTransport.peerCredentials['DN']
        return self.transformationDB.extendTransformation(transName, nTasks, author=authorDN)

    types_getTasksToSubmit = [transTypes, list(six.integer_types)]

    def export_getTasksToSubmit(self, transName, numTasks, site=""):
        """Get information necessary for submission for a given number of tasks for a given transformation"""
        res = self.transformationDB.getTransformation(transName)
        if not res["OK"]:
            return res
        transDict = res["Value"]
        submitDict = {}
        res = self.transformationDB.getTasksForSubmission(
            transName, numTasks=numTasks, site=site, statusList=["Created"]
        )
        if not res["OK"]:
            return res
        tasksDict = res["Value"]
        for taskID, taskDict in tasksDict.items():
            res = self.transformationDB.reserveTask(transName, int(taskID))
            if not res["OK"]:
                return res
            else:
                submitDict[taskID] = taskDict
        transDict["JobDictionary"] = submitDict
        return S_OK(transDict)

    ####################################################################
    #
    # These are the methods for TransformationMetaQueries table. It replaces methods
    # for the old TransformationInputDataQuery table
    #

    types_createTransformationMetaQuery = [transTypes, dict, six.string_types]

    def export_createTransformationMetaQuery(self, transName, queryDict, queryType):
        credDict = self.getRemoteCredentials()
        authorDN = credDict.get("DN", credDict.get("CN"))
        return self.transformationDB.createTransformationMetaQuery(transName, queryDict, queryType, author=authorDN)

    types_deleteTransformationMetaQuery = [transTypes, six.string_types]

    def export_deleteTransformationMetaQuery(self, transName, queryType):
        credDict = self.getRemoteCredentials()
        authorDN = credDict.get("DN", credDict.get("CN"))
        return self.transformationDB.deleteTransformationMetaQuery(transName, queryType, author=authorDN)

    types_getTransformationMetaQuery = [transTypes, six.string_types]

    def export_getTransformationMetaQuery(self, transName, queryType):
        return self.transformationDB.getTransformationMetaQuery(transName, queryType)

    ####################################################################
    #
    # These are the methods for transformation logging manipulation
    #

    types_getTransformationLogging = [transTypes]

    def export_getTransformationLogging(self, transName):
        return self.transformationDB.getTransformationLogging(transName)

    ####################################################################
    #
    # These are the methods for transformation additional parameters
    #

    types_getAdditionalParameters = [transTypes]

    def export_getAdditionalParameters(self, transName):
        return self.transformationDB.getAdditionalParameters(transName)

    ####################################################################
    #
    # These are the methods for file manipulation
    #

    types_getFileSummary = [list]

    @classmethod
    def export_getFileSummary(cls, lfns):
        return cls.transformationDB.getFileSummary(lfns)

    types_addDirectory = [six.string_types]

    @classmethod
    def export_addDirectory(cls, path, force=False):
        return cls.transformationDB.addDirectory(path, force=force)

    types_exists = [list]

    @classmethod
    def export_exists(cls, lfns):
        return cls.transformationDB.exists(lfns)

    types_addFile = [[list, dict, six.string_types]]

    @classmethod
    def export_addFile(cls, fileDicts, force=False):
        """Interface provides { LFN1 : { PFN1, SE1, ... }, LFN2 : { PFN2, SE2, ... } }"""
        return cls.transformationDB.addFile(fileDicts, force=force)

    types_removeFile = [[list, dict]]

    @classmethod
    def export_removeFile(cls, lfns):
        """Interface provides [ LFN1, LFN2, ... ]"""
        if isinstance(lfns, dict):
            lfns = list(lfns)
        return cls.transformationDB.removeFile(lfns)

    types_setMetadata = [six.string_types, dict]

    @classmethod
    def export_setMetadata(cls, path, querydict):
        """Set metadata to a file or to a directory (path)"""
        return cls.transformationDB.setMetadata(path, querydict)

    ####################################################################
    #
    # These are the methods used for web monitoring
    #

    # TODO Get rid of this (talk to Matvey)
    types_getDistinctAttributeValues = [six.string_types, dict]

    @classmethod
    def export_getDistinctAttributeValues(cls, attribute, selectDict):
        res = cls.transformationDB.getTableDistinctAttributeValues("Transformations", [attribute], selectDict)
        if not res["OK"]:
            return res
        return S_OK(res["Value"][attribute])

    types_getTableDistinctAttributeValues = [six.string_types, list, dict]

    @classmethod
    def export_getTableDistinctAttributeValues(cls, table, attributes, selectDict):
        return cls.transformationDB.getTableDistinctAttributeValues(table, attributes, selectDict)

    types_getTransformationStatusCounters = []

    @classmethod
    def export_getTransformationStatusCounters(cls):
        res = cls.transformationDB.getCounters("Transformations", ["Status"], {})
        if not res["OK"]:
            return res
        statDict = {}
        for attrDict, count in res["Value"]:
            statDict[attrDict["Status"]] = count
        return S_OK(statDict)

    types_getTransformationSummary = []

    def export_getTransformationSummary(self):
        """Get the summary of the currently existing transformations"""
        res = self.transformationDB.getTransformations()
        if not res["OK"]:
            return res
        transList = res["Value"]
        resultDict = {}
        for transDict in transList:
            transID = transDict["TransformationID"]
            res = self.transformationDB.getTransformationTaskStats(transID)
            if not res["OK"]:
                self.log.warn("Failed to get job statistics for transformation", transID)
                continue
            transDict["JobStats"] = res["Value"]
            res = self.transformationDB.getTransformationStats(transID)
            if not res["OK"]:
                transDict["NumberOfFiles"] = -1
            else:
                transDict["NumberOfFiles"] = res["Value"]["Total"]
            resultDict[transID] = transDict
        return S_OK(resultDict)

    types_getTabbedSummaryWeb = [six.string_types, dict, dict, list, int, int]

    def export_getTabbedSummaryWeb(self, table, requestedTables, selectDict, sortList, startItem, maxItems):
        tableDestinations = {
            "Transformations": {
                "TransformationFiles": ["TransformationID"],
                "TransformationTasks": ["TransformationID"],
            },
            "TransformationFiles": {
                "Transformations": ["TransformationID"],
                "TransformationTasks": ["TransformationID", "TaskID"],
            },
            "TransformationTasks": {
                "Transformations": ["TransformationID"],
                "TransformationFiles": ["TransformationID", "TaskID"],
            },
        }

        tableSelections = {
            "Transformations": ["TransformationID", "AgentType", "Type", "TransformationGroup", "Plugin"],
            "TransformationFiles": ["TransformationID", "TaskID", "Status", "UsedSE", "TargetSE"],
            "TransformationTasks": ["TransformationID", "TaskID", "ExternalStatus", "TargetSE"],
        }

        tableTimeStamps = {
            "Transformations": "CreationDate",
            "TransformationFiles": "LastUpdate",
            "TransformationTasks": "CreationTime",
        }

        tableStatusColumn = {
            "Transformations": "Status",
            "TransformationFiles": "Status",
            "TransformationTasks": "ExternalStatus",
        }

        resDict = {}
        res = self.__getTableSummaryWeb(
            table,
            selectDict,
            sortList,
            startItem,
            maxItems,
            selectColumns=tableSelections[table],
            timeStamp=tableTimeStamps[table],
            statusColumn=tableStatusColumn[table],
        )
        if not res["OK"]:
            self.log.error("Failed to get Summary for table", "%s %s" % (table, res["Message"]))
            return res
        resDict[table] = res["Value"]
        selections = res["Value"]["Selections"]
        tableSelection = {}
        for destination in tableDestinations[table].keys():
            tableSelection[destination] = {}
            for parameter in tableDestinations[table][destination]:
                tableSelection[destination][parameter] = selections.get(parameter, [])

        for table, paramDict in requestedTables.items():
            sortList = paramDict.get("SortList", [])
            startItem = paramDict.get("StartItem", 0)
            maxItems = paramDict.get("MaxItems", 50)
            res = self.__getTableSummaryWeb(
                table,
                tableSelection[table],
                sortList,
                startItem,
                maxItems,
                selectColumns=tableSelections[table],
                timeStamp=tableTimeStamps[table],
                statusColumn=tableStatusColumn[table],
            )
            if not res["OK"]:
                self.log.error("Failed to get Summary for table", "%s %s" % (table, res["Message"]))
                return res
            resDict[table] = res["Value"]
        return S_OK(resDict)

    types_getTransformationsSummaryWeb = [dict, list, int, int]

    def export_getTransformationsSummaryWeb(self, selectDict, sortList, startItem, maxItems):
        return self.__getTableSummaryWeb(
            "Transformations",
            selectDict,
            sortList,
            startItem,
            maxItems,
            selectColumns=["TransformationID", "AgentType", "Type", "Group", "Plugin"],
            timeStamp="CreationDate",
            statusColumn="Status",
        )

    types_getTransformationTasksSummaryWeb = [dict, list, int, int]

    def export_getTransformationTasksSummaryWeb(self, selectDict, sortList, startItem, maxItems):
        return self.__getTableSummaryWeb(
            "TransformationTasks",
            selectDict,
            sortList,
            startItem,
            maxItems,
            selectColumns=["TransformationID", "ExternalStatus", "TargetSE"],
            timeStamp="CreationTime",
            statusColumn="ExternalStatus",
        )

    types_getTransformationFilesSummaryWeb = [dict, list, int, int]

    def export_getTransformationFilesSummaryWeb(self, selectDict, sortList, startItem, maxItems):
        return self.__getTableSummaryWeb(
            "TransformationFiles",
            selectDict,
            sortList,
            startItem,
            maxItems,
            selectColumns=["TransformationID", "Status", "UsedSE", "TargetSE"],
            timeStamp="LastUpdate",
            statusColumn="Status",
        )

    def __getTableSummaryWeb(
        self, table, selectDict, sortList, startItem, maxItems, selectColumns=[], timeStamp=None, statusColumn="Status"
    ):
        fromDate = selectDict.get("FromDate", None)
        if fromDate:
            del selectDict["FromDate"]
        # if not fromDate:
        #  fromDate = last_update
        toDate = selectDict.get("ToDate", None)
        if toDate:
            del selectDict["ToDate"]
        # Sorting instructions. Only one for the moment.
        if sortList:
            orderAttribute = sortList[0][0] + ":" + sortList[0][1]
        else:
            orderAttribute = None
        # Get the columns that match the selection
        fcn = None
        fcnName = "get%s" % table
        if hasattr(self.transformationDB, fcnName) and callable(getattr(self.transformationDB, fcnName)):
            fcn = getattr(self.transformationDB, fcnName)
        if not fcn:
            return S_ERROR("Unable to invoke gTransformationDB.%s, it isn't a member function" % fcnName)
        res = fcn(condDict=selectDict, older=toDate, newer=fromDate, timeStamp=timeStamp, orderAttribute=orderAttribute)
        if not res["OK"]:
            return res

        # The full list of columns in contained here
        allRows = res["Records"]
        # Prepare the standard structure now within the resultDict dictionary
        resultDict = {}
        # Create the total records entry
        resultDict["TotalRecords"] = len(allRows)
        # Create the ParameterNames entry
        resultDict["ParameterNames"] = res["ParameterNames"]
        # Find which element in the tuple contains the requested status
        if statusColumn not in resultDict["ParameterNames"]:
            return S_ERROR("Provided status column not present")
        statusColumnIndex = resultDict["ParameterNames"].index(statusColumn)

        # Get the rows which are within the selected window
        if resultDict["TotalRecords"] == 0:
            return S_OK(resultDict)
        ini = startItem
        last = ini + maxItems
        if ini >= resultDict["TotalRecords"]:
            return S_ERROR("Item number out of range")
        if last > resultDict["TotalRecords"]:
            last = resultDict["TotalRecords"]
        selectedRows = allRows[ini:last]
        resultDict["Records"] = selectedRows

        # Generate the status dictionary
        statusDict = {}
        for row in selectedRows:
            status = row[statusColumnIndex]
            statusDict[status] = statusDict.setdefault(status, 0) + 1
        resultDict["Extras"] = statusDict

        # Obtain the distinct values of the selection parameters
        res = self.transformationDB.getTableDistinctAttributeValues(
            table, selectColumns, selectDict, older=toDate, newer=fromDate
        )
        distinctSelections = zip(selectColumns, [])
        if res["OK"]:
            distinctSelections = res["Value"]
        resultDict["Selections"] = distinctSelections

        return S_OK(resultDict)

    types_getTransformationSummaryWeb = [dict, list, int, int]

    def export_getTransformationSummaryWeb(self, selectDict, sortList, startItem, maxItems):
        """Get the summary of the transformation information for a given page in the generic format"""

        # Obtain the timing information from the selectDict
        last_update = selectDict.get("CreationDate", None)
        if last_update:
            del selectDict["CreationDate"]
        fromDate = selectDict.get("FromDate", None)
        if fromDate:
            del selectDict["FromDate"]
        if not fromDate:
            fromDate = last_update
        toDate = selectDict.get("ToDate", None)
        if toDate:
            del selectDict["ToDate"]
        # Sorting instructions. Only one for the moment.
        if sortList:
            orderAttribute = []
            for i in sortList:
                orderAttribute += [i[0] + ":" + i[1]]
        else:
            orderAttribute = None

        # Get the transformations that match the selection
        res = self.transformationDB.getTransformations(
            condDict=selectDict, older=toDate, newer=fromDate, orderAttribute=orderAttribute
        )
        if not res["OK"]:
            return res

        ops = Operations()
        # Prepare the standard structure now within the resultDict dictionary
        resultDict = {}
        trList = res["Records"]
        # Create the total records entry
        nTrans = len(trList)
        resultDict["TotalRecords"] = nTrans
        # Create the ParameterNames entry
        # As this list is a reference to the list in the DB, we cannot extend it, therefore copy it
        resultDict["ParameterNames"] = list(res["ParameterNames"])
        # Add the job states to the ParameterNames entry
        taskStateNames = TASKS_STATE_NAMES + ops.getValue("Transformations/AdditionalTaskStates", [])
        resultDict["ParameterNames"] += ["Jobs_" + x for x in taskStateNames]
        # Add the file states to the ParameterNames entry
        fileStateNames = FILES_STATE_NAMES + ops.getValue("Transformations/AdditionalFileStates", [])
        resultDict["ParameterNames"] += ["Files_" + x for x in fileStateNames]

        # Get the transformations which are within the selected window
        if nTrans == 0:
            return S_OK(resultDict)
        ini = startItem
        last = ini + maxItems
        if ini >= nTrans:
            return S_ERROR("Item number out of range")
        if last > nTrans:
            last = nTrans
        transList = trList[ini:last]

        statusDict = {}
        extendableTranfs = ops.getValue("Transformations/ExtendableTransfTypes", ["Simulation", "MCsimulation"])
        givenUpFileStatus = ops.getValue("Transformations/GivenUpFileStatus", ["MissingInFC"])
        problematicStatuses = ops.getValue("Transformations/ProblematicStatuses", ["Problematic"])
        # Add specific information for each selected transformation
        for trans in transList:
            transDict = dict(zip(resultDict["ParameterNames"], trans))

            # Update the status counters
            status = transDict["Status"]
            statusDict[status] = statusDict.setdefault(status, 0) + 1

            # Get the statistics on the number of jobs for the transformation
            transID = transDict["TransformationID"]
            res = self.transformationDB.getTransformationTaskStats(transID)
            taskDict = {}
            if res["OK"] and res["Value"]:
                taskDict = res["Value"]
            for state in taskStateNames:
                trans.append(taskDict.get(state, 0))

            # Get the statistics for the number of files for the transformation
            fileDict = {}
            transType = transDict["Type"]
            if transType.lower() in extendableTranfs:
                fileDict["PercentProcessed"] = "-"
            else:
                res = self.transformationDB.getTransformationStats(transID)
                if res["OK"]:
                    fileDict = res["Value"]
                    total = fileDict["Total"]
                    for stat in givenUpFileStatus:
                        total -= fileDict.get(stat, 0)
                    processed = fileDict.get(TransformationFilesStatus.PROCESSED, 0)
                    fileDict["PercentProcessed"] = "%.1f" % (int(processed * 1000.0 / total) / 10.0) if total else 0.0
            problematic = 0
            for stat in problematicStatuses:
                problematic += fileDict.get(stat, 0)
            fileDict["Problematic"] = problematic
            for state in fileStateNames:
                trans.append(fileDict.get(state, 0))

        resultDict["Records"] = transList
        resultDict["Extras"] = statusDict
        return S_OK(resultDict)

    ###########################################################################



"""TransformationCleaningAgent cleans up finalised transformations.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN TransformationCleaningAgent
  :end-before: ##END
  :dedent: 2
  :caption: TransformationCleaningAgent options

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

# # imports
import re
import ast
import os
import errno
from datetime import datetime, timedelta

# # from DIRAC
from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Core.Utilities.Proxy import executeWithUserProxy
from DIRAC.Core.Utilities.DErrno import cmpError
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient import JobMonitoringClient
from DIRAC.WorkloadManagementSystem.Client.WMSClient import WMSClient

# # agent's name
AGENT_NAME = "Transformation/TransformationCleaningAgent"


class TransformationCleaningAgent(AgentModule):
    """
    .. class:: TransformationCleaningAgent

    :param ~DIRAC.DataManagementSystem.Client.DataManager.DataManager dm: DataManager instance
    :param ~TransformationClient.TransformationClient transClient: TransformationClient instance
    :param ~FileCatalogClient.FileCatalogClient metadataClient: FileCatalogClient instance

    """

    def __init__(self, *args, **kwargs):
        """c'tor"""
        AgentModule.__init__(self, *args, **kwargs)

        self.shifterProxy = None

        # # transformation client
        self.transClient = None
        # # wms client
        self.wmsClient = None
        # # request client
        self.reqClient = None
        # # file catalog client
        self.metadataClient = None

        # # transformations types
        self.transformationTypes = None
        # # directory locations
        self.directoryLocations = ["TransformationDB", "MetadataCatalog"]
        # # transformation metadata
        self.transfidmeta = "TransformationID"
        # # archive periof in days
        self.archiveAfter = 7
        # # transformation log SEs
        self.logSE = "LogSE"
        # # enable/disable execution
        self.enableFlag = "True"

        self.dataProcTTypes = ["MCSimulation", "Merge"]
        self.dataManipTTypes = ["Replication", "Removal"]

    def initialize(self):
        """agent initialisation

        reading and setting config opts

        :param self: self reference
        """
        # # shifter proxy
        # See cleanContent method: this proxy will be used ALSO when the file catalog used
        # is the DIRAC File Catalog (DFC).
        # This is possible because of unset of the "UseServerCertificate" option
        self.shifterProxy = self.am_getOption("shifterProxy", self.shifterProxy)

        # # transformations types
        self.dataProcTTypes = Operations().getValue("Transformations/DataProcessing", self.dataProcTTypes)
        self.dataManipTTypes = Operations().getValue("Transformations/DataManipulation", self.dataManipTTypes)
        agentTSTypes = self.am_getOption("TransformationTypes", [])
        if agentTSTypes:
            self.transformationTypes = sorted(agentTSTypes)
        else:
            self.transformationTypes = sorted(self.dataProcTTypes + self.dataManipTTypes)
        self.log.info("Will consider the following transformation types: %s" % str(self.transformationTypes))
        # # directory locations
        self.directoryLocations = sorted(self.am_getOption("DirectoryLocations", self.directoryLocations))
        self.log.info("Will search for directories in the following locations: %s" % str(self.directoryLocations))
        # # transformation metadata
        self.transfidmeta = self.am_getOption("TransfIDMeta", self.transfidmeta)
        self.log.info("Will use %s as metadata tag name for TransformationID" % self.transfidmeta)
        # # archive periof in days
        self.archiveAfter = self.am_getOption("ArchiveAfter", self.archiveAfter)  # days
        self.log.info("Will archive Completed transformations after %d days" % self.archiveAfter)
        # # transformation log SEs
        self.logSE = Operations().getValue("/LogStorage/LogSE", self.logSE)
        self.log.info("Will remove logs found on storage element: %s" % self.logSE)

        # # transformation client
        self.transClient = TransformationClient()
        # # wms client
        self.wmsClient = WMSClient()
        # # request client
        self.reqClient = ReqClient()
        # # file catalog client
        self.metadataClient = FileCatalogClient()
        # # job monitoring client
        self.jobMonitoringClient = JobMonitoringClient()

        return S_OK()

    #############################################################################
    def execute(self):
        """execution in one agent's cycle

        :param self: self reference
        """

        self.enableFlag = self.am_getOption("EnableFlag", self.enableFlag)
        if self.enableFlag != "True":
            self.log.info("TransformationCleaningAgent is disabled by configuration option EnableFlag")
            return S_OK("Disabled via CS flag")

        # Obtain the transformations in Cleaning status and remove any mention of the jobs/files
        res = self.transClient.getTransformations({"Status": "Cleaning", "Type": self.transformationTypes})
        if res["OK"]:
            for transDict in res["Value"]:
                if self.shifterProxy:
                    self._executeClean(transDict)
                else:
                    self.log.info(
                        "Cleaning transformation %(TransformationID)s with %(AuthorDN)s, %(AuthorGroup)s" % transDict
                    )
                    executeWithUserProxy(self._executeClean)(
                        transDict, proxyUserDN=transDict["AuthorDN"], proxyUserGroup=transDict["AuthorGroup"]
                    )
        else:
            self.log.error("Failed to get transformations", res["Message"])

        # Obtain the transformations in RemovingFiles status and removes the output files
        res = self.transClient.getTransformations({"Status": "RemovingFiles", "Type": self.transformationTypes})
        if res["OK"]:
            for transDict in res["Value"]:
                if self.shifterProxy:
                    self._executeRemoval(transDict)
                else:
                    self.log.info(
                        "Removing files for transformation %(TransformationID)s with %(AuthorDN)s, %(AuthorGroup)s"
                        % transDict
                    )
                    executeWithUserProxy(self._executeRemoval)(
                        transDict, proxyUserDN=transDict["AuthorDN"], proxyUserGroup=transDict["AuthorGroup"]
                    )
        else:
            self.log.error("Could not get the transformations", res["Message"])

        # Obtain the transformations in Completed status and archive if inactive for X days
        olderThanTime = datetime.utcnow() - timedelta(days=self.archiveAfter)
        res = self.transClient.getTransformations(
            {"Status": "Completed", "Type": self.transformationTypes}, older=olderThanTime, timeStamp="LastUpdate"
        )
        if res["OK"]:
            for transDict in res["Value"]:
                if self.shifterProxy:
                    self._executeArchive(transDict)
                else:
                    self.log.info(
                        "Archiving files for transformation %(TransformationID)s with %(AuthorDN)s, %(AuthorGroup)s"
                        % transDict
                    )
                    executeWithUserProxy(self._executeArchive)(
                        transDict, proxyUserDN=transDict["AuthorDN"], proxyUserGroup=transDict["AuthorGroup"]
                    )
        else:
            self.log.error("Could not get the transformations", res["Message"])
        return S_OK()

    def finalize(self):
        """Only at finalization: will clean ancient transformations (remnants)

            1) get the transformation IDs of jobs that are older than 1 year
            2) find the status of those transformations. Those "Cleaned" and "Archived" will be
               cleaned and archived (again)

        Why doing this here? Basically, it's a race:

        1) the production manager submits a transformation
        2) the TransformationAgent, and a bit later the WorkflowTaskAgent, put such transformation in their internal queue,
           so eventually during their (long-ish) cycle they'll work on it.
        3) 1 minute after creating the transformation, the production manager cleans it (by hand, for whatever reason).
           So, the status is changed to "Cleaning"
        4) the TransformationCleaningAgent cleans what has been created (maybe, nothing),
           then sets the transformation status to "Cleaned" or "Archived"
        5) a bit later the TransformationAgent, and later the WorkflowTaskAgent, kick in,
           creating tasks and jobs for a production that's effectively cleaned (but these 2 agents don't know yet).

        Of course, one could make one final check in TransformationAgent or WorkflowTaskAgent,
        but these 2 agents are already doing a lot of stuff, and are pretty heavy.
        So, we should just clean from time to time.
        What I added here is done only when the agent finalize, and it's quite light-ish operation anyway.
        """
        res = self.jobMonitoringClient.getJobGroups(None, datetime.utcnow() - timedelta(days=365))
        if not res["OK"]:
            self.log.error("Failed to get job groups", res["Message"])
            return res
        transformationIDs = res["Value"]
        if transformationIDs:
            res = self.transClient.getTransformations({"TransformationID": transformationIDs})
            if not res["OK"]:
                self.log.error("Failed to get transformations", res["Message"])
                return res
            transformations = res["Value"]
            toClean = []
            toArchive = []
            for transDict in transformations:
                if transDict["Status"] == "Cleaned":
                    toClean.append(transDict)
                if transDict["Status"] == "Archived":
                    toArchive.append(transDict)

            for transDict in toClean:
                if self.shifterProxy:
                    self._executeClean(transDict)
                else:
                    self.log.info(
                        "Cleaning transformation %(TransformationID)s with %(AuthorDN)s, %(AuthorGroup)s" % transDict
                    )
                    executeWithUserProxy(self._executeClean)(
                        transDict, proxyUserDN=transDict["AuthorDN"], proxyUserGroup=transDict["AuthorGroup"]
                    )

            for transDict in toArchive:
                if self.shifterProxy:
                    self._executeArchive(transDict)
                else:
                    self.log.info(
                        "Archiving files for transformation %(TransformationID)s with %(AuthorDN)s, %(AuthorGroup)s"
                        % transDict
                    )
                    executeWithUserProxy(self._executeArchive)(
                        transDict, proxyUserDN=transDict["AuthorDN"], proxyUserGroup=transDict["AuthorGroup"]
                    )

            # Remove JobIDs that were unknown to the TransformationSystem
            jobGroupsToCheck = [str(transDict["TransformationID"]).zfill(8) for transDict in toClean + toArchive]
            res = self.jobMonitoringClient.getJobs({"JobGroup": jobGroupsToCheck})
            if not res["OK"]:
                return res
            jobIDsToRemove = [int(jobID) for jobID in res["Value"]]
            res = self.__removeWMSTasks(jobIDsToRemove)
            if not res["OK"]:
                return res

        return S_OK()

    def _executeClean(self, transDict):
        """Clean transformation."""
        # if transformation is of type `Replication` or `Removal`, there is nothing to clean.
        # We just archive
        if transDict["Type"] in self.dataManipTTypes:
            res = self.archiveTransformation(transDict["TransformationID"])
            if not res["OK"]:
                self.log.error(
                    "Problems archiving transformation", "%s: %s" % (transDict["TransformationID"], res["Message"])
                )
        else:
            res = self.cleanTransformation(transDict["TransformationID"])
            if not res["OK"]:
                self.log.error(
                    "Problems cleaning transformation", "%s: %s" % (transDict["TransformationID"], res["Message"])
                )

    def _executeRemoval(self, transDict):
        """Remove files from given transformation."""
        res = self.removeTransformationOutput(transDict["TransformationID"])
        if not res["OK"]:
            self.log.error(
                "Problems removing transformation", "%s: %s" % (transDict["TransformationID"], res["Message"])
            )

    def _executeArchive(self, transDict):
        """Archive the given transformation."""
        res = self.archiveTransformation(transDict["TransformationID"])
        if not res["OK"]:
            self.log.error(
                "Problems archiving transformation", "%s: %s" % (transDict["TransformationID"], res["Message"])
            )

        return S_OK()

    #############################################################################
    #
    # Get the transformation directories for checking
    #

    def getTransformationDirectories(self, transID):
        """get the directories for the supplied transformation from the transformation system.
            These directories are used by removeTransformationOutput and cleanTransformation for removing output.

        :param self: self reference
        :param int transID: transformation ID
        """
        self.log.verbose("Cleaning Transformation directories of transformation %d" % transID)
        directories = []
        if "TransformationDB" in self.directoryLocations:
            res = self.transClient.getTransformationParameters(transID, ["OutputDirectories"])
            if not res["OK"]:
                self.log.error("Failed to obtain transformation directories", res["Message"])
                return res
            transDirectories = []
            if res["Value"]:
                if not isinstance(res["Value"], list):
                    try:
                        transDirectories = ast.literal_eval(res["Value"])
                    except Exception:
                        # It can happen if the res['Value'] is '/a/b/c' instead of '["/a/b/c"]'
                        transDirectories.append(res["Value"])
                else:
                    transDirectories = res["Value"]
            directories = self._addDirs(transID, transDirectories, directories)

        if "MetadataCatalog" in self.directoryLocations:
            res = self.metadataClient.findDirectoriesByMetadata({self.transfidmeta: transID})
            if not res["OK"]:
                self.log.error("Failed to obtain metadata catalog directories", res["Message"])
                return res
            transDirectories = res["Value"]
            directories = self._addDirs(transID, transDirectories, directories)

        if not directories:
            self.log.info("No output directories found")
        directories = sorted(directories)
        return S_OK(directories)

    @classmethod
    def _addDirs(cls, transID, newDirs, existingDirs):
        """append unique :newDirs: list to :existingDirs: list

        :param self: self reference
        :param int transID: transformationID
        :param list newDirs: src list of paths
        :param list existingDirs: dest list of paths
        """
        for folder in newDirs:
            transStr = str(transID).zfill(8)
            if re.search(transStr, str(folder)):
                if folder not in existingDirs:
                    existingDirs.append(os.path.normpath(folder))
        return existingDirs

    #############################################################################
    #
    # These are the methods for performing the cleaning of catalogs and storage
    #

    def cleanContent(self, directory):
        """wipe out everything from catalog under folder :directory:

        :param self: self reference
        :params str directory: folder name
        """
        self.log.verbose("Cleaning Catalog contents")
        res = self.__getCatalogDirectoryContents([directory])
        if not res["OK"]:
            return res
        filesFound = res["Value"]
        if not filesFound:
            self.log.info("No files are registered in the catalog directory %s" % directory)
            return S_OK()
        self.log.info("Attempting to remove possible remnants from the catalog and storage", "(n=%d)" % len(filesFound))

        # Executing with shifter proxy
        gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "false")
        res = DataManager().removeFile(filesFound, force=True)
        gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "true")

        if not res["OK"]:
            return res
        realFailure = False
        for lfn, reason in res["Value"]["Failed"].items():
            if "File does not exist" in str(reason):
                self.log.warn("File %s not found in some catalog: " % (lfn))
            else:
                self.log.error("Failed to remove file found in the catalog", "%s %s" % (lfn, reason))
                realFailure = True
        if realFailure:
            return S_ERROR("Failed to remove all files found in the catalog")
        return S_OK()

    def __getCatalogDirectoryContents(self, directories):
        """get catalog contents under paths :directories:

        :param self: self reference
        :param list directories: list of paths in catalog
        """
        self.log.info("Obtaining the catalog contents for %d directories:" % len(directories))
        for directory in directories:
            self.log.info(directory)
        activeDirs = directories
        allFiles = {}
        fc = FileCatalog()
        while activeDirs:
            currentDir = activeDirs[0]
            res = returnSingleResult(fc.listDirectory(currentDir))
            activeDirs.remove(currentDir)
            if not res["OK"] and "Directory does not exist" in res["Message"]:  # FIXME: DFC should return errno
                self.log.info("The supplied directory %s does not exist" % currentDir)
            elif not res["OK"]:
                if "No such file or directory" in res["Message"]:
                    self.log.info("%s: %s" % (currentDir, res["Message"]))
                else:
                    self.log.error("Failed to get directory %s content" % currentDir, res["Message"])
            else:
                dirContents = res["Value"]
                activeDirs.extend(dirContents["SubDirs"])
                allFiles.update(dirContents["Files"])
        self.log.info("", "Found %d files" % len(allFiles))
        return S_OK(list(allFiles))

    def cleanTransformationLogFiles(self, directory):
        """clean up transformation logs from directory :directory:

        :param self: self reference
        :param str directory: folder name
        """
        self.log.verbose("Removing log files found in the directory", directory)
        res = returnSingleResult(StorageElement(self.logSE).removeDirectory(directory, recursive=True))
        if not res["OK"]:
            if cmpError(res, errno.ENOENT):  # No such file or directory
                self.log.warn("Transformation log directory does not exist", directory)
                return S_OK()
            self.log.error("Failed to remove log files", res["Message"])
            return res
        self.log.info("Successfully removed transformation log directory")
        return S_OK()

    #############################################################################
    #
    # These are the functional methods for archiving and cleaning transformations
    #

    def removeTransformationOutput(self, transID):
        """This just removes any mention of the output data from the catalog and storage"""
        self.log.info("Removing output data for transformation %s" % transID)
        res = self.getTransformationDirectories(transID)
        if not res["OK"]:
            self.log.error("Problem obtaining directories for transformation", "%s with result '%s'" % (transID, res))
            return S_OK()
        directories = res["Value"]
        for directory in directories:
            if not re.search("/LOG/", directory):
                res = self.cleanContent(directory)
                if not res["OK"]:
                    return res

        self.log.info(
            "Removed %d directories from the catalog \
      and its files from the storage for transformation %s"
            % (len(directories), transID)
        )
        # Clean ALL the possible remnants found in the metadata catalog
        res = self.cleanMetadataCatalogFiles(transID)
        if not res["OK"]:
            return res
        self.log.info("Successfully removed output of transformation", transID)
        # Change the status of the transformation to RemovedFiles
        res = self.transClient.setTransformationParameter(transID, "Status", "RemovedFiles")
        if not res["OK"]:
            self.log.error("Failed to update status of transformation %s to RemovedFiles" % (transID), res["Message"])
            return res
        self.log.info("Updated status of transformation %s to RemovedFiles" % (transID))
        return S_OK()

    def archiveTransformation(self, transID):
        """This just removes job from the jobDB and the transformation DB

        :param self: self reference
        :param int transID: transformation ID
        """
        self.log.info("Archiving transformation %s" % transID)
        # Clean the jobs in the WMS and any failover requests found
        res = self.cleanTransformationTasks(transID)
        if not res["OK"]:
            return res
        # Clean the transformation DB of the files and job information
        res = self.transClient.cleanTransformation(transID)
        if not res["OK"]:
            return res
        self.log.info("Successfully archived transformation %d" % transID)
        # Change the status of the transformation to archived
        res = self.transClient.setTransformationParameter(transID, "Status", "Archived")
        if not res["OK"]:
            self.log.error("Failed to update status of transformation %s to Archived" % (transID), res["Message"])
            return res
        self.log.info("Updated status of transformation %s to Archived" % (transID))
        return S_OK()

    def cleanTransformation(self, transID):
        """This removes what was produced by the supplied transformation,
        leaving only some info and log in the transformation DB.
        """
        self.log.info("Cleaning transformation", transID)
        res = self.getTransformationDirectories(transID)
        if not res["OK"]:
            self.log.error(
                "Problem obtaining directories for transformation", "%s with result '%s'" % (transID, res["Message"])
            )
            return S_OK()
        directories = res["Value"]
        # Clean the jobs in the WMS and any failover requests found
        res = self.cleanTransformationTasks(transID)
        if not res["OK"]:
            return res
        # Clean the log files for the jobs
        for directory in directories:
            if re.search("/LOG/", directory):
                res = self.cleanTransformationLogFiles(directory)
                if not res["OK"]:
                    return res
            res = self.cleanContent(directory)
            if not res["OK"]:
                return res

        # Clean ALL the possible remnants found
        res = self.cleanMetadataCatalogFiles(transID)
        if not res["OK"]:
            return res
        # Clean the transformation DB of the files and job information
        res = self.transClient.cleanTransformation(transID)
        if not res["OK"]:
            return res
        self.log.info("Successfully cleaned transformation", transID)
        res = self.transClient.setTransformationParameter(transID, "Status", "Cleaned")
        if not res["OK"]:
            self.log.error("Failed to update status of transformation %s to Cleaned" % (transID), res["Message"])
            return res
        self.log.info("Updated status of transformation", "%s to Cleaned" % (transID))
        return S_OK()

    def cleanMetadataCatalogFiles(self, transID):
        """wipe out files from catalog"""
        res = self.metadataClient.findFilesByMetadata({self.transfidmeta: transID})
        if not res["OK"]:
            return res
        fileToRemove = res["Value"]
        if not fileToRemove:
            self.log.info("No files found for transID", transID)
            return S_OK()

        # Executing with shifter proxy
        gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "false")
        res = DataManager().removeFile(fileToRemove, force=True)
        gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "true")

        if not res["OK"]:
            return res
        for lfn, reason in res["Value"]["Failed"].items():
            self.log.error("Failed to remove file found in metadata catalog", "%s %s" % (lfn, reason))
        if res["Value"]["Failed"]:
            return S_ERROR("Failed to remove all files found in the metadata catalog")
        self.log.info("Successfully removed all files found in the DFC")
        return S_OK()

    #############################################################################
    #
    # These are the methods for removing the jobs from the WMS and transformation DB
    #

    def cleanTransformationTasks(self, transID):
        """clean tasks from WMS, or from the RMS if it is a DataManipulation transformation"""
        self.log.verbose("Cleaning Transformation tasks of transformation", transID)
        res = self.__getTransformationExternalIDs(transID)
        if not res["OK"]:
            return res
        externalIDs = res["Value"]
        if externalIDs:
            res = self.transClient.getTransformationParameters(transID, ["Type"])
            if not res["OK"]:
                self.log.error("Failed to determine transformation type")
                return res
            transType = res["Value"]
            if transType in self.dataProcTTypes:
                res = self.__removeWMSTasks(externalIDs)
            else:
                res = self.__removeRequests(externalIDs)
            if not res["OK"]:
                return res
        return S_OK()

    def __getTransformationExternalIDs(self, transID):
        """collect all ExternalIDs for transformation :transID:

        :param self: self reference
        :param int transID: transforamtion ID
        """
        res = self.transClient.getTransformationTasks(condDict={"TransformationID": transID})
        if not res["OK"]:
            self.log.error("Failed to get externalIDs for transformation %d" % transID, res["Message"])
            return res
        externalIDs = [taskDict["ExternalID"] for taskDict in res["Value"]]
        self.log.info("Found %d tasks for transformation" % len(externalIDs))
        return S_OK(externalIDs)

    def __removeRequests(self, requestIDs):
        """This will remove requests from the RMS system -"""
        rIDs = [int(int(j)) for j in requestIDs if int(j)]
        for reqID in rIDs:
            self.reqClient.cancelRequest(reqID)

        return S_OK()

    def __removeWMSTasks(self, transJobIDs):
        """delete jobs (mark their status as "JobStatus.DELETED") and their requests from the system

        :param self: self reference
        :param list trasnJobIDs: job IDs
        """
        # Prevent 0 job IDs
        jobIDs = [int(j) for j in transJobIDs if int(j)]
        allRemove = True
        for jobList in breakListIntoChunks(jobIDs, 500):

            res = self.wmsClient.killJob(jobList)
            if res["OK"]:
                self.log.info("Successfully killed %d jobs from WMS" % len(jobList))
            elif ("InvalidJobIDs" in res) and ("NonauthorizedJobIDs" not in res) and ("FailedJobIDs" not in res):
                self.log.info("Found jobs which did not exist in the WMS", "(n=%d)" % len(res["InvalidJobIDs"]))
            elif "NonauthorizedJobIDs" in res:
                self.log.error("Failed to kill jobs because not authorized", "(n=%d)" % len(res["NonauthorizedJobIDs"]))
                allRemove = False
            elif "FailedJobIDs" in res:
                self.log.error("Failed to kill jobs", "(n=%d)" % len(res["FailedJobIDs"]))
                allRemove = False

            res = self.wmsClient.deleteJob(jobList)
            if res["OK"]:
                self.log.info("Successfully deleted jobs from WMS", "(n=%d)" % len(jobList))
            elif ("InvalidJobIDs" in res) and ("NonauthorizedJobIDs" not in res) and ("FailedJobIDs" not in res):
                self.log.info("Found jobs which did not exist in the WMS", "(n=%d)" % len(res["InvalidJobIDs"]))
            elif "NonauthorizedJobIDs" in res:
                self.log.error(
                    "Failed to delete jobs because not authorized", "(n=%d)" % len(res["NonauthorizedJobIDs"])
                )
                allRemove = False
            elif "FailedJobIDs" in res:
                self.log.error("Failed to delete jobs", "(n=%d)" % len(res["FailedJobIDs"]))
                allRemove = False

        if not allRemove:
            return S_ERROR("Failed to delete all remnants from WMS")
        self.log.info("Successfully deleted all tasks from the WMS")

        if not jobIDs:
            self.log.info("JobIDs not present, unable to delete associated requests.")
            return S_OK()

        failed = 0
        failoverRequests = {}
        res = self.reqClient.getRequestIDsForJobs(jobIDs)
        if not res["OK"]:
            self.log.error("Failed to get requestID for jobs.", res["Message"])
            return res
        failoverRequests.update(res["Value"]["Successful"])
        if not failoverRequests:
            return S_OK()
        for jobID, requestID in res["Value"]["Successful"].items():
            # Put this check just in case, tasks must have associated jobs
            if jobID == 0 or jobID == "0":
                continue
            res = self.reqClient.cancelRequest(requestID)
            if not res["OK"]:
                self.log.error("Failed to remove request from RequestDB", res["Message"])
                failed += 1
            else:
                self.log.verbose("Removed request %s associated to job %d." % (requestID, jobID))

        if failed:
            self.log.info("Successfully removed requests", "(n=%d)" % (len(failoverRequests) - failed))
            self.log.info("Failed to remove requests", "(n=%d)" % failed)
            return S_ERROR("Failed to remove all the request from RequestDB")
        self.log.info("Successfully removed all the associated failover requests")
        return S_OK()



"""
TornadoServer create a web server and load services.
It may work better with TornadoClient but as it accepts HTTPS you can create your own client
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import time
import datetime
import os
import six

import M2Crypto

import tornado.iostream

tornado.iostream.SSLIOStream.configure(
    "tornado_m2crypto.m2iostream.M2IOStream"
)  # pylint: disable=wrong-import-position

from tornado.httpserver import HTTPServer
from tornado.web import Application, url
from tornado.ioloop import IOLoop
import tornado.ioloop

import DIRAC
from DIRAC import gConfig, gLogger
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.Core.Security import Locations
from DIRAC.Core.Tornado.Server.HandlerManager import HandlerManager
from DIRAC.Core.Utilities import MemStat
from DIRAC.FrameworkSystem.Client.MonitoringClient import MonitoringClient

sLog = gLogger.getSubLogger(__name__)
DEBUG_M2CRYPTO = os.getenv("DIRAC_DEBUG_M2CRYPTO", "No").lower() in ("yes", "true")


class TornadoServer(object):
    """
    Tornado webserver

    Initialize and run an HTTPS Server for DIRAC services.
    By default it load all https services defined in the CS,
    but you can also give an explicit list.

    The listening port is either:

    * Given as parameter
    * Loaded from the CS ``/Systems/Tornado/<instance>/Port``
    * Default to 8443


    Example 1: Easy way to start tornado::

      # Initialize server and load services
      serverToLaunch = TornadoServer()

      # Start listening when ready
      serverToLaunch.startTornado()

    Example 2:We want to debug service1 and service2 only, and use another port for that ::

      services = ['component/service1', 'component/service2']
      serverToLaunch = TornadoServer(services=services, port=1234)
      serverToLaunch.startTornado()

    """

    def __init__(self, services=None, port=None):
        """

        :param list services: (default None) List of service handlers to load. If ``None``, loads all
        :param int port: Port to listen to. If None, the port is resolved following the logic
           described in the class documentation
        """

        if port is None:
            port = gConfig.getValue("/Systems/Tornado/%s/Port" % PathFinder.getSystemInstance("Tornado"), 8443)

        if services and not isinstance(services, list):
            services = [services]

        # URLs for services.
        # Contains Tornado :py:class:`tornado.web.url` object
        self.urls = []
        # Other infos
        self.port = port
        self.handlerManager = HandlerManager()

        # Monitoring attributes

        self._monitor = MonitoringClient()
        # temp value for computation, used by the monitoring
        self.__report = None
        # Last update time stamp
        self.__monitorLastStatsUpdate = None
        self.__monitoringLoopDelay = 60  # In secs

        # If services are defined, load only these ones (useful for debug purpose or specific services)
        if services:
            retVal = self.handlerManager.loadHandlersByServiceName(services)
            if not retVal["OK"]:
                sLog.error(retVal["Message"])
                raise ImportError("Some services can't be loaded, check the service names and configuration.")

        # if no service list is given, load services from configuration
        handlerDict = self.handlerManager.getHandlersDict()
        for item in handlerDict.items():
            # handlerDict[key].initializeService(key)
            self.urls.append(url(item[0], item[1]))
        # If there is no services loaded:
        if not self.urls:
            raise ImportError("There is no services loaded, please check your configuration")

    def startTornado(self):
        """
        Starts the tornado server when ready.
        This method never returns.
        """

        sLog.debug("Starting Tornado")
        self._initMonitoring()

        router = Application(self.urls, debug=False, compress_response=True)

        certs = Locations.getHostCertificateAndKeyLocation()
        if certs is False:
            sLog.fatal("Host certificates not found ! Can't start the Server")
            raise ImportError("Unable to load certificates")
        ca = Locations.getCAsLocation()
        ssl_options = {
            "certfile": certs[0],
            "keyfile": certs[1],
            "cert_reqs": M2Crypto.SSL.verify_peer,
            "ca_certs": ca,
            "sslDebug": DEBUG_M2CRYPTO,  # Set to true if you want to see the TLS debug messages
        }

        self.__monitorLastStatsUpdate = time.time()
        self.__report = self.__startReportToMonitoringLoop()

        # Starting monitoring, IOLoop waiting time in ms, __monitoringLoopDelay is defined in seconds
        tornado.ioloop.PeriodicCallback(self.__reportToMonitoring, self.__monitoringLoopDelay * 1000).start()

        # If we are running with python3, Tornado will use asyncio,
        # and we have to convince it to let us run in a different thread
        # Doing this ensures a consistent behavior between py2 and py3
        if six.PY3:
            import asyncio  # pylint: disable=import-error

            asyncio.set_event_loop_policy(tornado.platform.asyncio.AnyThreadEventLoopPolicy())

        # Start server
        server = HTTPServer(router, ssl_options=ssl_options, decompress_request=True)
        try:
            server.listen(self.port)
        except Exception as e:  # pylint: disable=broad-except
            sLog.exception("Exception starting HTTPServer", e)
            raise
        sLog.always("Listening on port %s" % self.port)
        for service in self.urls:
            sLog.debug("Available service: %s" % service)

        IOLoop.current().start()

    def _initMonitoring(self):
        """
        Initialize the monitoring
        """

        self._monitor.setComponentType(MonitoringClient.COMPONENT_TORNADO)
        self._monitor.initialize()
        self._monitor.setComponentName("Tornado")

        self._monitor.registerActivity("CPU", "CPU Usage", "Framework", "CPU,%", MonitoringClient.OP_MEAN, 600)
        self._monitor.registerActivity("MEM", "Memory Usage", "Framework", "Memory,MB", MonitoringClient.OP_MEAN, 600)

        self._monitor.setComponentExtraParam("DIRACVersion", DIRAC.version)
        self._monitor.setComponentExtraParam("platform", DIRAC.getPlatform())
        self._monitor.setComponentExtraParam("startTime", datetime.datetime.utcnow())

    def __reportToMonitoring(self):
        """
        Periodically report to the monitoring of the CPU and MEM
        """

        # Calculate CPU usage by comparing realtime and cpu time since last report
        self.__endReportToMonitoringLoop(*self.__report)

        # Save memory usage and save realtime/CPU time for next call
        self.__report = self.__startReportToMonitoringLoop()

    def __startReportToMonitoringLoop(self):
        """
        Snapshot of resources to be taken at the beginning
        of a monitoring cycle.
        Also sends memory snapshot to the monitoring.

        This is basically copy/paste of Service.py

        :returns: tuple (<time.time(), cpuTime )

        """
        now = time.time()  # Used to calulate a delta
        stats = os.times()
        cpuTime = stats[0] + stats[2]
        if now - self.__monitorLastStatsUpdate < 0:
            return (now, cpuTime)
        # Send CPU consumption mark
        self.__monitorLastStatsUpdate = now
        # Send Memory consumption mark
        membytes = MemStat.VmB("VmRSS:")
        if membytes:
            mem = membytes / (1024.0 * 1024.0)
            self._monitor.addMark("MEM", mem)
        return (now, cpuTime)

    def __endReportToMonitoringLoop(self, initialWallTime, initialCPUTime):
        """
        Snapshot of resources to be taken at the end
        of a monitoring cycle.

        This is basically copy/paste of Service.py

        Determines CPU usage by comparing walltime and cputime and send it to monitor
        """
        wallTime = time.time() - initialWallTime
        stats = os.times()
        cpuTime = stats[0] + stats[2] - initialCPUTime
        percentage = cpuTime / wallTime * 100.0
        if percentage > 0:
            self._monitor.addMark("CPU", percentage)



########################################################################
# File: RequestDB.py
# pylint: disable=no-member
########################################################################

""" Frontend for ReqDB

    :mod: RequestDB

    =======================

    .. module: RequestDB

    :synopsis: db holding Requests

    db holding Request, Operation and File
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
import errno
import random

import datetime

from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm import relationship, backref, sessionmaker, joinedload, mapper
from sqlalchemy.sql import update
from sqlalchemy import (
    create_engine,
    func,
    Table,
    Column,
    MetaData,
    ForeignKey,
    Integer,
    String,
    DateTime,
    Enum,
    BLOB,
    BigInteger,
    distinct,
)

# # from DIRAC
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File
from DIRAC.ConfigurationSystem.Client.Utilities import getDBParameters


__RCSID__ = "$Id$"


# Metadata instance that is used to bind the engine, Object and tables
metadata = MetaData()


# Description of the file table

fileTable = Table(
    "File",
    metadata,
    Column("FileID", Integer, primary_key=True),
    Column("OperationID", Integer, ForeignKey("Operation.OperationID", ondelete="CASCADE"), nullable=False),
    Column("Status", Enum("Waiting", "Done", "Failed", "Scheduled"), server_default="Waiting"),
    Column("LFN", String(255), index=True),
    Column("PFN", String(255)),
    Column("ChecksumType", Enum("ADLER32", "MD5", "SHA1", ""), server_default=""),
    Column("Checksum", String(255)),
    Column("GUID", String(36)),
    Column("Size", BigInteger),
    Column("Attempt", Integer),
    Column("Error", String(255)),
    mysql_engine="InnoDB",
)

# Map the File object to the fileTable, with a few special attributes

mapper(
    File,
    fileTable,
    properties={
        "_Status": fileTable.c.Status,
        "_LFN": fileTable.c.LFN,
        "_ChecksumType": fileTable.c.ChecksumType,
        "_GUID": fileTable.c.GUID,
    },
)


# Description of the Operation table

operationTable = Table(
    "Operation",
    metadata,
    Column("TargetSE", String(255)),
    Column("CreationTime", DateTime),
    Column("SourceSE", String(255)),
    Column("Arguments", BLOB),
    Column("Error", String(255)),
    Column("Type", String(64), nullable=False),
    Column("Order", Integer, nullable=False),
    Column(
        "Status",
        Enum("Waiting", "Assigned", "Queued", "Done", "Failed", "Canceled", "Scheduled"),
        server_default="Queued",
    ),
    Column("LastUpdate", DateTime),
    Column("SubmitTime", DateTime),
    Column("Catalog", String(255)),
    Column("OperationID", Integer, primary_key=True),
    Column("RequestID", Integer, ForeignKey("Request.RequestID", ondelete="CASCADE"), nullable=False),
    mysql_engine="InnoDB",
)


# Map the Operation object to the operationTable, with a few special attributes

mapper(
    Operation,
    operationTable,
    properties={
        "_CreationTime": operationTable.c.CreationTime,
        "_Arguments": operationTable.c.Arguments,
        "_Order": operationTable.c.Order,
        "_Status": operationTable.c.Status,
        "_LastUpdate": operationTable.c.LastUpdate,
        "_SubmitTime": operationTable.c.SubmitTime,
        "_Catalog": operationTable.c.Catalog,
        "__files__": relationship(
            File,
            backref=backref("_parent", lazy="immediate"),
            lazy="immediate",
            passive_deletes=True,
            cascade="all, delete-orphan",
        ),
    },
)


# Description of the Request Table

requestTable = Table(
    "Request",
    metadata,
    Column("DIRACSetup", String(32)),
    Column("CreationTime", DateTime),
    Column("JobID", Integer, server_default="0"),
    Column("OwnerDN", String(255)),
    Column("RequestName", String(255), nullable=False),
    Column("Error", String(255)),
    Column("Status", Enum("Waiting", "Assigned", "Done", "Failed", "Canceled", "Scheduled"), server_default="Waiting"),
    Column("LastUpdate", DateTime),
    Column("OwnerGroup", String(32)),
    Column("SubmitTime", DateTime),
    Column("RequestID", Integer, primary_key=True),
    Column("SourceComponent", BLOB),
    Column("NotBefore", DateTime),
    mysql_engine="InnoDB",
)

# Map the Request object to the requestTable, with a few special attributes
mapper(
    Request,
    requestTable,
    properties={
        "_CreationTime": requestTable.c.CreationTime,
        "_SourceComponent": requestTable.c.SourceComponent,
        "_Status": requestTable.c.Status,
        "_LastUpdate": requestTable.c.LastUpdate,
        "_SubmitTime": requestTable.c.SubmitTime,
        "_NotBefore": requestTable.c.NotBefore,
        "__operations__": relationship(
            Operation,
            backref=backref("_parent", lazy="immediate"),
            order_by=operationTable.c.Order,
            lazy="immediate",
            passive_deletes=True,
            cascade="all, delete-orphan",
        ),
    },
)


########################################################################
class RequestDB(object):
    """
    .. class:: RequestDB

    db holding requests
    """

    def __getDBConnectionInfo(self, fullname):
        """Collect from the CS all the info needed to connect to the DB.
        This should be in a base class eventually
        """

        result = getDBParameters(fullname)
        if not result["OK"]:
            raise Exception("Cannot get database parameters: %s" % result["Message"])

        dbParameters = result["Value"]
        self.dbHost = dbParameters["Host"]
        self.dbPort = dbParameters["Port"]
        self.dbUser = dbParameters["User"]
        self.dbPass = dbParameters["Password"]
        self.dbName = dbParameters["DBName"]

    def __init__(self):
        """c'tor

        :param self: self reference
        """

        self.log = gLogger.getSubLogger("RequestDB")
        # Initialize the connection info
        self.__getDBConnectionInfo("RequestManagement/ReqDB")

        runDebug = gLogger.getLevel() == "DEBUG"
        self.engine = create_engine(
            "mysql://%s:%s@%s:%s/%s" % (self.dbUser, self.dbPass, self.dbHost, self.dbPort, self.dbName),
            echo=runDebug,
            pool_recycle=3600,
        )

        metadata.bind = self.engine

        self.DBSession = sessionmaker(bind=self.engine)

    def createTables(self):
        """create tables"""
        try:
            metadata.create_all(self.engine)
        except Exception as e:
            return S_ERROR(e)
        return S_OK()

    def cancelRequest(self, requestID):
        session = self.DBSession()
        try:
            updateRet = session.execute(
                update(Request)
                .where(Request.RequestID == requestID)
                .values({Request._Status: "Canceled", Request._LastUpdate: datetime.datetime.utcnow()})
                .execution_options(synchronize_session=False)
            )  # See FTS3DB for synchronize_session
            session.commit()

            # No row was changed
            if not updateRet.rowcount:
                return S_ERROR("No such request %s" % requestID)

            return S_OK()

        except Exception as e:
            session.rollback()
            self.log.exception("cancelRequest: unexpected exception", lException=e)
            return S_ERROR("cancelRequest: unexpected exception %s" % e)
        finally:
            session.close()

    def putRequest(self, request):
        """update or insert request into db

        :param ~Request.Request request: Request instance
        """

        session = self.DBSession(expire_on_commit=False)
        try:

            try:
                if hasattr(request, "RequestID"):

                    status = session.query(Request._Status).filter(Request.RequestID == request.RequestID).one()

                    if status[0] == "Canceled":
                        self.log.info(
                            "Request %s(%s) was canceled, don't put it back" % (request.RequestID, request.RequestName)
                        )
                        return S_OK(request.RequestID)

            except NoResultFound:
                pass

            # Since the object request is not attached to the session, we merge it to have an update
            # instead of an insert with duplicate primary key
            request = session.merge(request)
            session.add(request)
            session.commit()
            session.expunge_all()

            return S_OK(request.RequestID)

        except Exception as e:
            session.rollback()
            self.log.exception("putRequest: unexpected exception", lException=e)
            return S_ERROR("putRequest: unexpected exception %s" % e)
        finally:
            session.close()

    def getScheduledRequest(self, operationID):
        session = self.DBSession()
        try:
            requestID = (
                session.query(Request.RequestID)
                .join(Request.__operations__)
                .filter(Operation.OperationID == operationID)
                .one()
            )
            return self.getRequest(requestID[0])
        except NoResultFound:
            return S_OK()
        finally:
            session.close()

    #
    #   def getRequestName( self, requestID ):
    #     """ get Request.RequestName for a given Request.RequestID """
    #
    #     session = self.DBSession()
    #     try:
    #       requestName = session.query( Request.RequestName )\
    #                            .filter( Request.RequestID == requestID )\
    #                            .one()
    #       return S_OK( requestName[0] )
    #     except NoResultFound, e:
    #       return S_ERROR( "getRequestName: no request found for RequestID=%s" % requestID )
    #     finally:
    #       session.close()

    def getRequest(self, reqID=0, assigned=True):
        """read request for execution

        :param reqID: request's ID (default 0) If 0, take a pseudo random one

        """

        # expire_on_commit is set to False so that we can still use the object after we close the session
        session = self.DBSession(expire_on_commit=False)
        log = self.log.getSubLogger("getRequest" if assigned else "peekRequest")

        requestID = None

        try:

            if reqID:
                requestID = reqID

                log.verbose("selecting request '%s'%s" % (reqID, " (Assigned)" if assigned else ""))
                status = None
                try:
                    status = session.query(Request._Status).filter(Request.RequestID == reqID).one()
                except NoResultFound:
                    return S_ERROR("getRequest: request '%s' not exists" % reqID)

                if status and status == "Assigned" and assigned:
                    return S_ERROR(
                        "getRequest: status of request '%s' is 'Assigned', request cannot be selected" % reqID
                    )

            else:
                now = datetime.datetime.utcnow().replace(microsecond=0)
                reqIDs = set()
                try:
                    reqAscIDs = (
                        session.query(Request.RequestID)
                        .filter(Request._Status == "Waiting")
                        .filter(Request._NotBefore < now)
                        .order_by(Request._LastUpdate)
                        .limit(100)
                        .all()
                    )

                    reqIDs = set([reqID[0] for reqID in reqAscIDs])

                    reqDescIDs = (
                        session.query(Request.RequestID)
                        .filter(Request._Status == "Waiting")
                        .filter(Request._NotBefore < now)
                        .order_by(Request._LastUpdate.desc())
                        .limit(50)
                        .all()
                    )

                    reqIDs |= set([reqID[0] for reqID in reqDescIDs])
                # No Waiting requests
                except NoResultFound:
                    return S_OK()

                if not reqIDs:
                    return S_OK()

                reqIDs = list(reqIDs)
                random.shuffle(reqIDs)
                requestID = reqIDs[0]

            # If we are here, the request MUST exist, so no try catch
            # the joinedload is to force the non-lazy loading of all the attributes, especially _parent
            request = (
                session.query(Request)
                .options(joinedload("__operations__").joinedload("__files__"))
                .filter(Request.RequestID == requestID)
                .one()
            )

            if not reqID:
                log.verbose(
                    "selected request %s('%s')%s"
                    % (request.RequestID, request.RequestName, " (Assigned)" if assigned else "")
                )

            if assigned:
                session.execute(
                    update(Request)
                    .where(Request.RequestID == requestID)
                    .values({Request._Status: "Assigned", Request._LastUpdate: datetime.datetime.utcnow()})
                )
                session.commit()

            session.expunge_all()
            return S_OK(request)

        except Exception as e:
            session.rollback()
            log.exception("getRequest: unexpected exception", lException=e)
            return S_ERROR("getRequest: unexpected exception : %s" % e)
        finally:
            session.close()

    def getBulkRequests(self, numberOfRequest=10, assigned=True):
        """read as many requests as requested for execution

        :param int numberOfRequest: Number of Request we want (default 10)
        :param bool assigned: if True, the status of the selected requests are set to assign

        :returns: a dictionary of Request objects indexed on the RequestID

        """

        # expire_on_commit is set to False so that we can still use the object after we close the session
        session = self.DBSession(expire_on_commit=False)
        log = self.log.getSubLogger("getBulkRequest" if assigned else "peekBulkRequest")

        requestDict = {}

        try:
            # If we are here, the request MUST exist, so no try catch
            # the joinedload is to force the non-lazy loading of all the attributes, especially _parent
            try:
                now = datetime.datetime.utcnow().replace(microsecond=0)
                requestIDs = (
                    session.query(Request.RequestID)
                    .with_for_update()
                    .filter(Request._Status == "Waiting")
                    .filter(Request._NotBefore < now)
                    .order_by(Request._LastUpdate)
                    .limit(numberOfRequest)
                    .all()
                )

                requestIDs = [ridTuple[0] for ridTuple in requestIDs]
                log.debug("Got request ids %s" % requestIDs)

                requests = (
                    session.query(Request)
                    .options(joinedload("__operations__").joinedload("__files__"))
                    .filter(Request.RequestID.in_(requestIDs))
                    .all()
                )
                log.debug("Got %s Request objects " % len(requests))
                requestDict = dict((req.RequestID, req) for req in requests)
            # No Waiting requests
            except NoResultFound:
                pass

            if assigned and requestDict:
                session.execute(
                    update(Request)
                    .where(Request.RequestID.in_(requestDict.keys()))
                    .values({Request._Status: "Assigned", Request._LastUpdate: datetime.datetime.utcnow()})
                )
            session.commit()

            session.expunge_all()

        except Exception as e:
            session.rollback()
            log.exception("unexpected exception", lException=e)
            return S_ERROR("getBulkRequest: unexpected exception : %s" % e)
        finally:
            session.close()

        return S_OK(requestDict)

    def peekRequest(self, requestID):
        """get request (ro), no update on states

        :param requestID: Request.RequestID
        """
        return self.getRequest(requestID, False)

    def getRequestIDsList(self, statusList=None, limit=None, since=None, until=None, getJobID=False):
        """select requests with status in :statusList:"""
        statusList = statusList if statusList else list(Request.FINAL_STATES)
        limit = limit if limit else 100
        session = self.DBSession()
        requestIDs = []
        try:
            if getJobID:
                reqQuery = session.query(Request.RequestID, Request._Status, Request._LastUpdate, Request.JobID).filter(
                    Request._Status.in_(statusList)
                )
            else:
                reqQuery = session.query(Request.RequestID, Request._Status, Request._LastUpdate).filter(
                    Request._Status.in_(statusList)
                )
            if since:
                reqQuery = reqQuery.filter(Request._LastUpdate > since)
            if until:
                reqQuery = reqQuery.filter(Request._LastUpdate < until)

            reqQuery = reqQuery.order_by(Request._LastUpdate).limit(limit)
            requestIDs = [list(reqIDTuple) for reqIDTuple in reqQuery.all()]

        except Exception as e:
            session.rollback()
            self.log.exception("getRequestIDsList: unexpected exception", lException=e)
            return S_ERROR("getRequestIDsList: unexpected exception : %s" % e)
        finally:
            session.close()

        return S_OK(requestIDs)

    def deleteRequest(self, requestID):
        """delete request given its ID

        :param str requestID: request.RequestID
        :param mixed connection: connection to use if any
        """

        session = self.DBSession()

        try:
            session.query(Request).filter(Request.RequestID == requestID).delete()
            session.commit()
        except Exception as e:
            session.rollback()
            self.log.exception("deleteRequest: unexpected exception", lException=e)
            return S_ERROR("deleteRequest: unexpected exception : %s" % e)
        finally:
            session.close()

        return S_OK()

    def getDBSummary(self):
        """get db summary"""
        # # this will be returned
        retDict = {"Request": {}, "Operation": {}, "File": {}}

        session = self.DBSession()

        try:
            requestQuery = session.query(Request._Status, func.count(Request.RequestID)).group_by(Request._Status).all()

            for status, count in requestQuery:
                retDict["Request"][status] = count

            operationQuery = (
                session.query(Operation.Type, Operation._Status, func.count(Operation.OperationID))
                .group_by(Operation.Type, Operation._Status)
                .all()
            )

            for oType, status, count in operationQuery:
                retDict["Operation"].setdefault(oType, {})[status] = count

            fileQuery = session.query(File._Status, func.count(File.FileID)).group_by(File._Status).all()

            for status, count in fileQuery:
                retDict["File"][status] = count

        except Exception as e:
            self.log.exception("getDBSummary: unexpected exception", lException=e)
            return S_ERROR("getDBSummary: unexpected exception : %s" % e)
        finally:
            session.close()

        return S_OK(retDict)

    def getRequestSummaryWeb(self, selectDict, sortList, startItem, maxItems):
        """Returns a list of Request for the web portal

        :param dict selectDict: parameter on which to restrain the query {key : Value}
                                key can be any of the Request columns, 'Type' (interpreted as Operation.Type)
                                and 'FromData' and 'ToData' are matched against the LastUpdate field
        :param sortList: [sorting column, ASC/DESC]
        :type sortList: python:list
        :param int startItem: start item (for pagination)
        :param int maxItems: max items (for pagination)
        """

        parameterList = [
            "RequestID",
            "RequestName",
            "JobID",
            "OwnerDN",
            "OwnerGroup",
            "Status",
            "Error",
            "CreationTime",
            "LastUpdate",
        ]

        resultDict = {}

        session = self.DBSession()

        try:
            summaryQuery = session.query(
                Request.RequestID,
                Request.RequestName,
                Request.JobID,
                Request.OwnerDN,
                Request.OwnerGroup,
                Request._Status,
                Request.Error,
                Request._CreationTime,
                Request._LastUpdate,
            )

            for key, value in selectDict.items():

                if key == "ToDate":
                    summaryQuery = summaryQuery.filter(Request._LastUpdate < value)
                elif key == "FromDate":
                    summaryQuery = summaryQuery.filter(Request._LastUpdate > value)
                else:

                    tableName = "Request"

                    if key == "Type":
                        summaryQuery = summaryQuery.join(Request.__operations__).group_by(
                            Request.RequestID,
                            Request.RequestName,
                            Request.JobID,
                            Request.OwnerDN,
                            Request.OwnerGroup,
                            Request._Status,
                            Request.Error,
                            Request._CreationTime,
                            Request._LastUpdate,
                            Operation.Type,
                        )
                        tableName = "Operation"
                    elif key == "Status":
                        key = "_Status"

                    if isinstance(value, list):
                        summaryQuery = summaryQuery.filter(eval("%s.%s.in_(%s)" % (tableName, key, value)))
                    else:
                        summaryQuery = summaryQuery.filter(eval("%s.%s" % (tableName, key)) == value)

            if sortList:
                summaryQuery = summaryQuery.order_by(eval("Request.%s.%s()" % (sortList[0][0], sortList[0][1].lower())))

            try:
                requestLists = summaryQuery.all()
            except NoResultFound:
                resultDict["ParameterNames"] = parameterList
                resultDict["Records"] = []

                return S_OK(resultDict)
            except Exception as e:
                return S_ERROR("Error getting the webSummary %s" % e)

            nRequests = len(requestLists)

            if startItem <= len(requestLists):
                firstIndex = startItem
            else:
                return S_ERROR("getRequestSummaryWeb: Requested index out of range")

            if (startItem + maxItems) <= len(requestLists):
                secondIndex = startItem + maxItems
            else:
                secondIndex = len(requestLists)

            records = []
            for i in range(firstIndex, secondIndex):
                row = requestLists[i]
                records.append([str(x) for x in row])

            resultDict["ParameterNames"] = parameterList
            resultDict["Records"] = records
            resultDict["TotalRecords"] = nRequests

            return S_OK(resultDict)
        #
        except Exception as e:
            self.log.exception("getRequestSummaryWeb: unexpected exception", lException=e)
            return S_ERROR("getRequestSummaryWeb: unexpected exception : %s" % e)

        finally:
            session.close()

    def getRequestCountersWeb(self, groupingAttribute, selectDict):
        """For the web portal.
        Returns a dictionary {value : counts} for a given key.
        The key can be any field from the RequestTable. or "Type",
        which will be interpreted as 'Operation.Type'
        """

        resultDict = {}

        session = self.DBSession()

        if groupingAttribute == "Type":
            groupingAttribute = "Operation.Type"
        elif groupingAttribute == "Status":
            groupingAttribute = "Request._Status"
        else:
            groupingAttribute = "Request.%s" % groupingAttribute

        try:
            summaryQuery = session.query(eval(groupingAttribute), func.count(Request.RequestID))

            for key, value in selectDict.items():
                if key == "ToDate":
                    summaryQuery = summaryQuery.filter(Request._LastUpdate < value)
                elif key == "FromDate":
                    summaryQuery = summaryQuery.filter(Request._LastUpdate > value)
                else:

                    objectType = "Request"
                    if key == "Type":
                        summaryQuery = summaryQuery.join(Request.__operations__)
                        objectType = "Operation"
                    elif key == "Status":
                        key = "_Status"

                    if isinstance(value, list):
                        summaryQuery = summaryQuery.filter(eval("%s.%s.in_(%s)" % (objectType, key, value)))
                    else:
                        summaryQuery = summaryQuery.filter(eval("%s.%s" % (objectType, key)) == value)

            summaryQuery = summaryQuery.group_by(eval(groupingAttribute))

            try:
                requestLists = summaryQuery.all()
                resultDict = dict(requestLists)
            except NoResultFound:
                pass
            except Exception as e:
                return S_ERROR("Error getting the webCounters %s" % e)

            return S_OK(resultDict)

        except Exception as e:
            self.log.exception("getRequestSummaryWeb: unexpected exception", lException=e)
            return S_ERROR("getRequestSummaryWeb: unexpected exception : %s" % e)

        finally:
            session.close()

    def getDistinctValues(self, tableName, columnName):
        """For a given table and a given field, return the list of of distinct values in the DB"""

        session = self.DBSession()
        distinctValues = []
        if columnName == "Status":
            columnName = "_Status"
        try:
            result = session.query(distinct(eval("%s.%s" % (tableName, columnName)))).all()
            distinctValues = [dist[0] for dist in result]
        except NoResultFound:
            pass
        except Exception as e:
            self.log.exception("getDistinctValues: unexpected exception", lException=e)
            return S_ERROR("getDistinctValues: unexpected exception : %s" % e)

        finally:
            session.close()

        return S_OK(distinctValues)

    def getRequestIDsForJobs(self, jobIDs):
        """returns request ids for jobs given jobIDs

        :param list jobIDs: list of jobIDs
        :return: S_OK( "Successful" : { jobID1 : Request, jobID2: Request, ... }
                       "Failed" : { jobID3: "error message", ... } )
        """
        self.log.debug("getRequestIDsForJobs: got %s jobIDs to check" % str(jobIDs))
        if not jobIDs:
            return S_ERROR("Must provide jobID list as argument.")
        if isinstance(jobIDs, six.integer_types):
            jobIDs = [jobIDs]
        jobIDs = set(jobIDs)

        reqDict = {"Successful": {}, "Failed": {}}

        session = self.DBSession()

        try:
            ret = session.query(Request.JobID, Request.RequestID).filter(Request.JobID.in_(jobIDs)).all()

            reqDict["Successful"] = dict((jobId, reqID) for jobId, reqID in ret)
            reqDict["Failed"] = dict((jobid, "Request not found") for jobid in jobIDs - set(reqDict["Successful"]))
        except Exception as e:
            self.log.exception("getRequestIDsForJobs: unexpected exception", lException=e)
            return S_ERROR("getRequestIDsForJobs: unexpected exception : %s" % e)
        finally:
            session.close()

        return S_OK(reqDict)

    def readRequestsForJobs(self, jobIDs=None):
        """read request for jobs

        :param list jobIDs: list of JobIDs
        :return: S_OK( "Successful" : { jobID1 : Request, jobID2: Request, ... }
                       "Failed" : { jobID3: "error message", ... } )
        """
        self.log.debug("readRequestForJobs: got %s jobIDs to check" % str(jobIDs))
        if not jobIDs:
            return S_ERROR("Must provide jobID list as argument.")
        if isinstance(jobIDs, six.integer_types):
            jobIDs = [jobIDs]
        jobIDs = set(jobIDs)

        reqDict = {"Successful": {}, "Failed": {}}

        # expire_on_commit is set to False so that we can still use the object after we close the session
        session = self.DBSession(expire_on_commit=False)

        try:
            ret = (
                session.query(Request.JobID, Request)
                .options(joinedload("__operations__").joinedload("__files__"))
                .filter(Request.JobID.in_(jobIDs))
                .all()
            )

            reqDict["Successful"] = dict((jobId, reqObj) for jobId, reqObj in ret)

            reqDict["Failed"] = dict((jobid, "Request not found") for jobid in jobIDs - set(reqDict["Successful"]))
            session.expunge_all()
        except Exception as e:
            self.log.exception("readRequestsForJobs: unexpected exception", lException=e)
            return S_ERROR("readRequestsForJobs: unexpected exception : %s" % e)
        finally:
            session.close()

        return S_OK(reqDict)

    def getRequestStatus(self, requestID):
        """get request status for a given request ID"""
        self.log.debug("getRequestStatus: checking status for '%s' request" % requestID)
        session = self.DBSession()
        try:
            status = session.query(Request._Status).filter(Request.RequestID == requestID).one()
        except NoResultFound:
            return S_ERROR(errno.ENOENT, "Request %s does not exist" % requestID)
        finally:
            session.close()
        return S_OK(status[0])

    def getRequestFileStatus(self, requestID, lfnList):
        """get status for files in request given its id

        A single status is returned by file, which corresponds
        to the most representative one. That is:

        * Failed: if it has failed in any of the operation
        * Scheduled: if it is Scheduled in any of the operation
        * Waiting: if the process is ongoing
        * Done: if everything was executed

        :param str requestID: Request.RequestID
        :param lfnList: list of LFNs
        :type lfnList: python:list
        """

        session = self.DBSession()
        try:
            res = dict.fromkeys(lfnList, "UNKNOWN")
            requestRet = (
                session.query(File._LFN, File._Status)
                .join(Request.__operations__)
                .join(Operation.__files__)
                .filter(Request.RequestID == requestID)
                .filter(File._LFN.in_(lfnList))
                .order_by(Operation._Order)
                .all()
            )

            for lfn, status in requestRet:
                # If the file was in one of these two state in the previous
                # operations, that's the one we want to return
                if res.get(lfn) not in ("Failed", "Scheduled"):
                    res[lfn] = status
            return S_OK(res)

        except Exception as e:
            self.log.exception("getRequestFileStatus: unexpected exception", lException=e)
            return S_ERROR("getRequestFileStatus: unexpected exception : %s" % e)
        finally:
            session.close()

    def getRequestInfo(self, requestID):
        """get request info given Request.RequestID"""

        session = self.DBSession()

        try:

            requestInfoQuery = session.query(
                Request.RequestID,
                Request._Status,
                Request.RequestName,
                Request.JobID,
                Request.OwnerDN,
                Request.OwnerGroup,
                Request.DIRACSetup,
                Request._SourceComponent,
                Request._CreationTime,
                Request._SubmitTime,
                Request._LastUpdate,
            ).filter(Request.RequestID == requestID)

            try:
                requestInfo = requestInfoQuery.one()
            except NoResultFound:
                return S_ERROR("No such request")

            return S_OK(list(requestInfo))

        except Exception as e:
            self.log.exception("getRequestInfo: unexpected exception", lException=e)
            return S_ERROR("getRequestInfo: unexpected exception : %s" % e)

        finally:
            session.close()

    def getDigest(self, requestID):
        """get digest for request given its id

        :param str requestName: request id
        """
        self.log.debug("getDigest: will create digest for request '%s'" % requestID)
        request = self.getRequest(requestID, False)
        if not request["OK"]:
            self.log.error("getDigest: %s" % request["Message"])
            return request
        request = request["Value"]
        if not isinstance(request, Request):
            self.log.info("getDigest: request '%s' not found")
            return S_OK()
        return request.getDigest()

    def getRequestIDForName(self, requestName):
        """read request id for given name
            if the name is not unique, an error is returned

        :param requestName: name of the request
        """
        session = self.DBSession()

        reqID = 0
        try:
            ret = session.query(Request.RequestID).filter(Request.RequestName == requestName).all()
            if not ret:
                return S_ERROR("No such request %s" % requestName)
            elif len(ret) > 1:
                return S_ERROR("RequestName %s not unique (%s matches)" % (requestName, len(ret)))

            reqID = ret[0][0]

        except NoResultFound:
            return S_ERROR("No such request")
        except Exception as e:
            self.log.exception("getRequestIDsForName: unexpected exception", lException=e)
            return S_ERROR("getRequestIDsForName: unexpected exception : %s" % e)
        finally:
            session.close()

        return S_OK(reqID)



""" I wish who wrote this would have put some doc...

    IIUC this is a wrapper around the JobState object. It basically tries to cache
    everything locally instead of going to the DB.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
import copy
import time

from DIRAC.Core.Utilities import Time, DEncode
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.WorkloadManagementSystem.Client.JobState.JobState import JobState
from DIRAC.WorkloadManagementSystem.Client.JobState.JobManifest import JobManifest


class CachedJobState(object):

    log = gLogger.getSubLogger("CachedJobState")

    def __init__(self, jid, skipInitState=False):
        self.dOnlyCache = False
        self.__jid = jid
        self.__jobState = JobState(jid)
        self.cleanState(skipInitState=skipInitState)

    def cleanState(self, skipInitState=False):
        self.__cache = {}
        self.__jobLog = []
        self.__insertIntoTQ = False
        self.__dirtyKeys = set()
        self.__manifest = False
        self.__initState = None
        self.__lastValidState = time.time()
        if not skipInitState:
            result = self.getAttributes(["Status", "MinorStatus", "LastUpdateTime"])
            if result["OK"]:
                self.__initState = result["Value"]
            else:
                self.__initState = None

    def recheckValidity(self, graceTime=600):
        now = time.time()
        if graceTime <= 0 or now - self.__lastValidState > graceTime:
            self.__lastValidState = now
            result = self.__jobState.getAttributes(["Status", "MinorStatus", "LastUpdateTime"])
            if not result["OK"]:
                return result
            currentState = result["Value"]
            if not currentState == self.__initState:
                return S_OK(False)
            return S_OK(True)
        return S_OK(self.valid)

    @property
    def valid(self):
        return self.__initState is not None

    @property
    def jid(self):
        return self.__jid

    def getDirtyKeys(self):
        return set(self.__dirtyKeys)

    def commitChanges(self):
        if self.__initState is None:
            return S_ERROR("CachedJobState( %d ) is not valid" % self.__jid)
        changes = {}
        for k in self.__dirtyKeys:
            changes[k] = self.__cache[k]
        result = self.__jobState.commitCache(self.__initState, changes, self.__jobLog)
        try:
            result.pop("rpcStub")
        except KeyError:
            pass
        if not result["OK"]:
            self.cleanState()
            return result
        if not result["Value"]:
            self.cleanState()
            return S_ERROR("Initial state was different")
        newState = result["Value"]
        self.__jobLog = []
        self.__dirtyKeys.clear()
        # Save manifest
        if self.__manifest and self.__manifest.isDirty():
            result = self.__jobState.setManifest(self.__manifest)
            if not result["OK"]:
                self.cleanState()
                for _ in range(5):
                    if self.__jobState.rescheduleJob()["OK"]:
                        break
                return result
            self.__manifest.clearDirty()
        # Insert into TQ
        if self.__insertIntoTQ:
            result = self.__jobState.insertIntoTQ()
            if not result["OK"]:
                self.cleanState()
                for _ in range(5):
                    if self.__jobState.rescheduleJob()["OK"]:
                        break
                return result
            self.__insertIntoTQ = False

        self.__initState = newState
        self.__lastValidState = time.time()
        return S_OK()

    def serialize(self):
        if self.__manifest:
            manifest = [self.__manifest.dumpAsCFG(), self.__manifest.isDirty()]
        else:
            manifest = None
        data = DEncode.encode(
            [
                self.__jid,
                self.__cache,
                self.__jobLog,
                manifest,
                self.__initState,
                self.__insertIntoTQ,
                list(self.__dirtyKeys),
            ]
        )
        return data.decode()

    @staticmethod
    def deserialize(stub):
        dataTuple, _slen = DEncode.decode(stub.encode())
        if len(dataTuple) != 7:
            return S_ERROR("Invalid stub")
        # jid
        if not isinstance(dataTuple[0], six.integer_types):
            return S_ERROR("Invalid stub 0")
        # cache
        if not isinstance(dataTuple[1], dict):
            return S_ERROR("Invalid stub 1")
        # trace
        if not isinstance(dataTuple[2], list):
            return S_ERROR("Invalid stub 2")
        # manifest
        if dataTuple[3] is not None and (not isinstance(dataTuple[3], (tuple, list)) and len(dataTuple[3]) != 2):
            return S_ERROR("Invalid stub 3")
        # initstate
        if not isinstance(dataTuple[4], dict):
            return S_ERROR("Invalid stub 4")
        # Insert into TQ
        if not isinstance(dataTuple[5], bool):
            return S_ERROR("Invalid stub 5")
        # Dirty Keys
        if not isinstance(dataTuple[6], (tuple, list)):
            return S_ERROR("Invalid stub 6")
        cjs = CachedJobState(dataTuple[0], skipInitState=True)
        cjs.__cache = dataTuple[1]
        cjs.__jobLog = dataTuple[2]
        dt3 = dataTuple[3]
        if dataTuple[3]:
            manifest = JobManifest()
            result = manifest.loadCFG(dt3[0])
            if not result["OK"]:
                return result
            if dt3[1]:
                manifest.setDirty()
            else:
                manifest.clearDirty()
            cjs.__manifest = manifest
        cjs.__initState = dataTuple[4]
        cjs.__insertIntoTQ = dataTuple[5]
        cjs.__dirtyKeys = set(dataTuple[6])
        return S_OK(cjs)

    def __cacheAdd(self, key, value):
        self.__cache[key] = value
        self.__dirtyKeys.add(key)

    def __cacheExists(self, keyList):
        if isinstance(keyList, six.string_types):
            keyList = [keyList]
        for key in keyList:
            if key not in self.__cache:
                return False
        return True

    def __cacheResult(self, cKey, functor, fArgs=None):
        # If it's a string
        if isinstance(cKey, six.string_types):
            if cKey not in self.__cache:
                if self.dOnlyCache:
                    return S_ERROR("%s is not cached")
                if not fArgs:
                    fArgs = tuple()
                result = functor(*fArgs)
                if not result["OK"]:
                    return result
                data = result["Value"]
                self.__cache[cKey] = data
            return S_OK(self.__cache[cKey])
        # Tuple/List
        elif isinstance(cKey, (list, tuple)):
            if not self.__cacheExists(cKey):
                if self.dOnlyCache:
                    return S_ERROR("%s is not cached")
                if not fArgs:
                    fArgs = tuple()
                result = functor(*fArgs)
                if not result["OK"]:
                    return result
                data = result["Value"]
                if len(cKey) != len(data):
                    gLogger.warn(
                        "CachedJobState.__memorize( %s, %s = %s ) doesn't receive the same amount of values as keys"
                        % (cKey, functor, data)
                    )
                    return data
                for i, val in enumerate(cKey):
                    self.__cache[val] = data[i]
            # Prepare result
            return S_OK(tuple([self.__cache[cK] for cK in cKey]))
        else:
            raise RuntimeError("Cache key %s does not have a valid type" % cKey)

    def __cacheDict(self, prefix, functor, keyList=None):
        if not keyList or not self.__cacheExists(["%s.%s" % (prefix, key) for key in keyList]):
            result = functor(keyList)
            if not result["OK"]:
                return result
            data = result["Value"]
            for key in data:
                cKey = "%s.%s" % (prefix, key)
                # If the key is already in the cache. DO NOT TOUCH. User may have already modified it.
                # We update the coming data with the cached data
                if cKey in self.__cache:
                    data[key] = self.__cache[cKey]
                else:
                    self.__cache[cKey] = data[key]
            return S_OK(data)
        return S_OK(dict([(key, self.__cache["%s.%s" % (prefix, key)]) for key in keyList]))

    def _inspectCache(self):
        return copy.deepcopy(self.__cache)

    def _clearCache(self):
        self.__cache = {}

    @property
    def _internals(self):
        if self.__manifest:
            manifest = (self.__manifest.dumpAsCFG(), self.__manifest.isDirty())
        else:
            manifest = None
        return (
            self.__jid,
            self.dOnlyCache,
            dict(self.__cache),
            list(self.__jobLog),
            manifest,
            dict(self.__initState),
            list(self.__dirtyKeys),
        )

    #
    # Manifest
    #

    def getManifest(self):
        if not self.__manifest:
            result = self.__jobState.getManifest()
            if not result["OK"]:
                return result
            self.__manifest = result["Value"]
        return S_OK(self.__manifest)

    def setManifest(self, manifest):
        if not isinstance(manifest, JobManifest):
            jobManifest = JobManifest()
            result = jobManifest.load(str(manifest))
            if not result["OK"]:
                return result
            manifest = jobManifest
        manifest.setDirty()
        self.__manifest = manifest
        # self.__manifest.clearDirty()
        return S_OK()

    # Attributes
    #

    def __addLogRecord(self, majorStatus=None, minorStatus=None, appStatus=None, source=None):
        record = {}
        if majorStatus:
            record["status"] = majorStatus
        if minorStatus:
            record["minor"] = minorStatus
        if appStatus:
            record["application"] = appStatus
        if not record:
            return
        if not source:
            source = "Unknown"
        self.__jobLog.append([record, Time.dateTime(), source])

    def setStatus(self, majorStatus, minorStatus=None, appStatus=None, source=None):
        self.__cacheAdd("att.Status", majorStatus)
        if minorStatus:
            self.__cacheAdd("att.MinorStatus", minorStatus)
        if appStatus:
            self.__cacheAdd("att.ApplicationStatus", appStatus)
        self.__addLogRecord(majorStatus, minorStatus, appStatus, source)
        return S_OK()

    def setMinorStatus(self, minorStatus, source=None):
        self.__cacheAdd("att.MinorStatus", minorStatus)
        self.__addLogRecord(minorStatus=minorStatus, source=source)
        return S_OK()

    def getStatus(self):
        return self.__cacheResult(("att.Status", "att.MinorStatus"), self.__jobState.getStatus)

    def setAppStatus(self, appStatus, source=None):
        self.__cacheAdd("att.ApplicationStatus", appStatus)
        self.__addLogRecord(appStatus=appStatus, source=source)
        return S_OK()

    def getAppStatus(self):
        return self.__cacheResult("att.ApplicationStatus", self.__jobState.getAppStatus)

    #
    # Attribs
    #

    def setAttribute(self, name, value):
        if not isinstance(name, six.string_types):
            return S_ERROR("Attribute name has to be a string")
        self.__cacheAdd("att.%s" % name, value)
        return S_OK()

    def setAttributes(self, attDict):
        if not isinstance(attDict, dict):
            return S_ERROR("Attributes has to be a dictionary and it's %s" % str(type(attDict)))
        for key in attDict:
            self.__cacheAdd("att.%s" % key, attDict[key])
        return S_OK()

    def getAttribute(self, name):
        return self.__cacheResult("att.%s" % name, self.__jobState.getAttribute, (name,))

    def getAttributes(self, nameList=None):
        return self.__cacheDict("att", self.__jobState.getAttributes, nameList)

    # JobParameters --- REMOVED

    # Optimizer params

    def setOptParameter(self, name, value):
        if not isinstance(name, six.string_types):
            return S_ERROR("Optimizer parameter name has to be a string")
        self.__cacheAdd("optp.%s" % name, value)
        return S_OK()

    def setOptParameters(self, pDict):
        if not isinstance(pDict, dict):
            return S_ERROR("Optimizer parameters has to be a dictionary")
        for key in pDict:
            self.__cacheAdd("optp.%s" % key, pDict[key])
        return S_OK()

    def getOptParameter(self, name):
        return self.__cacheResult("optp.%s" % name, self.__jobState.getOptParameter, (name,))

    def getOptParameters(self, nameList=None):
        return self.__cacheDict("optp", self.__jobState.getOptParameters, nameList)

    # Other

    def resetJob(self, source=""):
        """Reset the job!"""
        return self.__jobState.resetJob(source=source)

    def getInputData(self):
        return self.__cacheResult("inputData", self.__jobState.getInputData)

    def insertIntoTQ(self):
        if self.valid:
            self.__insertIntoTQ = True
            return S_OK()
        return S_ERROR("Cached state is invalid")



""" DIRAC FileCatalog plugin class to manage file metadata
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.Time import queryTime
from DIRAC.Core.Utilities.List import intListToString
from DIRAC.DataManagementSystem.Client.MetaQuery import (
    FILE_STANDARD_METAKEYS,
    FILES_TABLE_METAKEYS,
    FILEINFO_TABLE_METAKEYS,
)


class FileMetadata(object):
    def __init__(self, database=None):

        self.db = database

    def setDatabase(self, database):
        self.db = database

    ##############################################################################
    #
    #  Manage Metadata fields
    #
    ##############################################################################
    def addMetadataField(self, pName, pType, credDict):
        """Add a new metadata parameter to the Metadata Database.

        :param str pName: parameter name
        :param str pType: parameter type in the MySQL notation

        :return: S_OK/S_ERROR, Value - comment on a positive result
        """

        if pName in FILE_STANDARD_METAKEYS:
            return S_ERROR("Illegal use of reserved metafield name")

        result = self.db.dmeta._getMetadataFields(credDict)
        if not result["OK"]:
            return result
        if pName in result["Value"]:
            return S_ERROR("The metadata %s is already defined for Directories" % pName)
        result = self._getFileMetadataFields(credDict)
        if not result["OK"]:
            return result
        if pName in result["Value"]:
            if pType.lower() == result["Value"][pName].lower():
                return S_OK("Already exists")
            else:
                return S_ERROR(
                    "Attempt to add an existing metadata with different type: %s/%s" % (pType, result["Value"][pName])
                )
        valueType = pType
        if pType == "MetaSet":
            valueType = "VARCHAR(64)"
        req = (
            "CREATE TABLE FC_FileMeta_%s ( FileID INTEGER NOT NULL, Value %s, PRIMARY KEY (FileID), INDEX (Value) )"
            % (pName, valueType)
        )
        result = self.db._query(req)
        if not result["OK"]:
            return result

        result = self.db.insertFields("FC_FileMetaFields", ["MetaName", "MetaType"], [pName, pType])
        if not result["OK"]:
            return result

        metadataID = result["lastRowId"]
        result = self.__transformMetaParameterToData(pName)
        if not result["OK"]:
            return result
        return S_OK("Added new metadata: %d" % metadataID)

    def deleteMetadataField(self, pName, credDict):
        """Remove metadata field

        :param str pName: meta parameter name
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR
        """

        req = "DROP TABLE FC_FileMeta_%s" % pName
        result = self.db._update(req)
        error = ""
        if not result["OK"]:
            error = result["Message"]
        req = "DELETE FROM FC_FileMetaFields WHERE MetaName='%s'" % pName
        result = self.db._update(req)
        if not result["OK"]:
            if error:
                result["Message"] = error + "; " + result["Message"]
        return result

    def getFileMetadataFields(self, credDict):
        """Get all the defined metadata fields

        :param dict credDict: client credential dictionary
        :return: S_OK/S_ERROR, Value is the metadata:metadata type dictionary
        """

        return self._getFileMetadataFields(credDict)

    def _getFileMetadataFields(self, credDict):
        """Get all the defined metadata fields as they are stored in the database

        :param dict credDict: client credential dictionary
        :return: S_OK/S_ERROR, Value is the metadata:metadata type dictionary
        """

        req = "SELECT MetaName,MetaType FROM FC_FileMetaFields"
        result = self.db._query(req)
        if not result["OK"]:
            return result

        metaDict = {}
        for row in result["Value"]:
            metaDict[row[0]] = row[1]

        return S_OK(metaDict)

    ###########################################################
    #
    # Set and get metadata for files
    #
    ###########################################################

    def setMetadata(self, path, metaDict, credDict):
        """Set the value of a given metadata field for the the given directory path

        :param str path: file path
        :param dict metaDict: dictionary with metadata
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR
        """
        result = self._getFileMetadataFields(credDict)
        if not result["OK"]:
            return result
        metaFields = result["Value"]

        result = self.db.fileManager._findFiles([path])
        if not result["OK"]:
            return result
        if result["Value"]["Successful"]:
            fileID = result["Value"]["Successful"][path]["FileID"]
        else:
            return S_ERROR("File %s not found" % path)

        for metaName, metaValue in metaDict.items():
            if metaName not in metaFields:
                result = self.__setFileMetaParameter(fileID, metaName, metaValue, credDict)
            else:
                result = self.db.insertFields("FC_FileMeta_%s" % metaName, ["FileID", "Value"], [fileID, metaValue])
                if not result["OK"]:
                    if result["Message"].find("Duplicate") != -1:
                        req = "UPDATE FC_FileMeta_%s SET Value='%s' WHERE FileID=%d" % (metaName, metaValue, fileID)
                        result = self.db._update(req)
                        if not result["OK"]:
                            return result
                    else:
                        return result

        return S_OK()

    def removeMetadata(self, path, metadata, credDict):
        """Remove the specified metadata for the given file

        :param str path: file path
        :param list metadata: list of meta parameter names
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR
        """
        result = self._getFileMetadataFields(credDict)
        if not result["OK"]:
            return result
        metaFields = result["Value"]

        result = self.db.fileManager._findFiles([path])
        if not result["OK"]:
            return result
        if result["Value"]["Successful"]:
            fileID = result["Value"]["Successful"][path]["FileID"]
        else:
            return S_ERROR("File %s not found" % path)

        failedMeta = {}
        for meta in metadata:
            if meta in metaFields:
                # Indexed meta case
                req = "DELETE FROM FC_FileMeta_%s WHERE FileID=%d" % (meta, fileID)
                result = self.db._update(req)
                if not result["OK"]:
                    failedMeta[meta] = result["Value"]
            else:
                # Meta parameter case
                req = "DELETE FROM FC_FileMeta WHERE MetaKey='%s' AND FileID=%d" % (meta, fileID)
                result = self.db._update(req)
                if not result["OK"]:
                    failedMeta[meta] = result["Value"]

        if failedMeta:
            metaExample = list(failedMeta)[0]
            result = S_ERROR("Failed to remove %d metadata, e.g. %s" % (len(failedMeta), failedMeta[metaExample]))
            result["FailedMetadata"] = failedMeta
        else:
            return S_OK()

    def __getFileID(self, path):
        """Get file ID for the given file name

        :param str path: file path

        :return: S_OK/S_ERROR, Value - file ID
        """

        result = self.db.fileManager._findFiles([path])
        if not result["OK"]:
            return result
        if result["Value"]["Successful"]:
            fileID = result["Value"]["Successful"][path]["FileID"]
        else:
            return S_ERROR("File not found")
        return S_OK(fileID)

    def __setFileMetaParameter(self, fileID, metaName, metaValue, credDict):
        """Set an meta parameter - metadata which is not used in the the data
        search operations

        :param int fileID: file ID
        :param str metaName: meta parameter name
        :param str metaValue: meta parameter value
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR
        """
        result = self.db.insertFields(
            "FC_FileMeta", ["FileID", "MetaKey", "MetaValue"], [fileID, metaName, str(metaValue)]
        )
        return result

    def setFileMetaParameter(self, path, metaName, metaValue, credDict):
        """Set an meta parameter - metadata which is not used in the the data
        search operations

        :param str path: file name
        :param str metaName: meta parameter name
        :param str metaValue: meta parameter value
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR
        """

        result = self.__getFileID(path)
        if not result["OK"]:
            return result
        fileID = result["Value"]
        return self.__setFileMetaParameter(fileID, metaName, metaValue, credDict)

    def _getFileUserMetadataByID(self, fileIDList, credDict, connection=False):
        """Get file user metadata for the list of file IDs

        :param list fileIDList: list of file IDs (int)
        :param dict credDict: client credential dictionary
        :param obj connection: database connection object

        :return: S_OK/S_ERROR, Value - dict of metadata for each input file
        """
        # First file metadata
        result = self._getFileMetadataFields(credDict)
        if not result["OK"]:
            return result
        metaFields = result["Value"]

        stringIDs = ",".join(["%s" % fId for fId in fileIDList])
        metaDict = {}
        for meta in metaFields:
            req = "SELECT Value,FileID FROM FC_FileMeta_%s WHERE FileID in (%s)" % (meta, stringIDs)
            result = self.db._query(req, conn=connection)
            if not result["OK"]:
                return result
            for value, fileID in result["Value"]:
                metaDict.setdefault(fileID, {})
                metaDict[fileID][meta] = value

        req = "SELECT FileID,MetaKey,MetaValue from FC_FileMeta where FileID in (%s)" % stringIDs
        result = self.db._query(req, conn=connection)
        if not result["OK"]:
            return result
        for fileID, key, value in result["Value"]:
            metaDict.setdefault(fileID, {})
            metaDict[fileID][key] = value

        return S_OK(metaDict)

    def getFileUserMetadata(self, path, credDict):
        """Get metadata for the given file

        :param str path: file name
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR, Value - dict of metadata
        """
        # First file metadata
        result = self._getFileMetadataFields(credDict)
        if not result["OK"]:
            return result
        metaFields = result["Value"]

        result = self.__getFileID(path)
        if not result["OK"]:
            return result
        fileID = result["Value"]

        metaDict = {}
        metaTypeDict = {}
        for meta in metaFields:
            req = "SELECT Value,FileID FROM FC_FileMeta_%s WHERE FileID=%d" % (meta, fileID)
            result = self.db._query(req)
            if not result["OK"]:
                return result
            if result["Value"]:
                metaDict[meta] = result["Value"][0][0]
            metaTypeDict[meta] = metaFields[meta]

        result = self.getFileMetaParameters(path, credDict)
        if result["OK"]:
            metaDict.update(result["Value"])
            for meta in result["Value"]:
                metaTypeDict[meta] = "NonSearchable"

        result = S_OK(metaDict)
        result["MetadataType"] = metaTypeDict
        return result

    def __getFileMetaParameters(self, fileID, credDict):
        """Get meta parameters for the given file

        :param int fileID: file ID
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR, Value - dict of meta parameters
        """

        req = "SELECT FileID,MetaKey,MetaValue from FC_FileMeta where FileID=%d " % fileID
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK({})
        metaDict = {}
        for fileID, key, value in result["Value"]:
            if key in metaDict:
                if isinstance(metaDict[key], list):
                    metaDict[key].append(value)
                else:
                    metaDict[key] = [metaDict[key]].append(value)
            else:
                metaDict[key] = value

        return S_OK(metaDict)

    def getFileMetaParameters(self, path, credDict):
        """Get meta parameters for the given file

        :param str path: file name
        :param dict credDict: client credential dictionary

        :return: S_OK/S_ERROR, Value - dict of meta parameters
        """

        result = self.__getFileID(path)
        if not result["OK"]:
            return result
        fileID = result["Value"]

        return self.__getFileMetaParameters(fileID, credDict)

    def __transformMetaParameterToData(self, metaName):
        """Relocate the meta parameters of all the files to the corresponding
        indexed metadata table

        :param str metaName: meta parameter name

        :return: S_OK/S_ERROR
        """

        req = "SELECT FileID,MetaValue from FC_FileMeta WHERE MetaKey='%s'" % metaName
        result = self.db._query(req)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK()

        insertValueList = []
        for fileID, meta in result["Value"]:
            insertValueList.append("( %d,'%s' )" % (fileID, meta))

        req = "INSERT INTO FC_FileMeta_%s (FileID,Value) VALUES %s" % (metaName, ", ".join(insertValueList))
        result = self.db._update(req)
        if not result["OK"]:
            return result

        req = "DELETE FROM FC_FileMeta WHERE MetaKey='%s'" % metaName
        result = self.db._update(req)
        return result

    #########################################################################
    #
    #  Finding files by metadata
    #
    #########################################################################

    def __createMetaSelection(self, value):
        """Create selection string for the given meta value to be used in the SQL query

        :param float,int,str value: metadata value

        :return: S_OK/S_ERROR, Value - list of SQL query selection elements
        """
        queryList = []
        if isinstance(value, float):
            queryList.append(("=", "%f" % value))
        elif isinstance(value, six.integer_types):
            queryList.append(("=", "%d" % value))
        elif isinstance(value, six.string_types):
            if value.lower() == "any":
                queryList.append(("IS", "NOT NULL"))
            elif value.lower() == "missing":
                queryList.append(("IS", "NULL"))
            elif value:
                result = self.db._escapeString(value)
                if not result["OK"]:
                    return result
                eValue = result["Value"]
                if "*" in eValue or "?" in eValue:
                    eValue = eValue.replace("*", "%%")
                    eValue = eValue.replace("?", "_")
                    queryList.append(("LIKE", eValue))
                else:
                    queryList.append(("=", eValue))
            else:
                queryList.append(("", ""))
        elif isinstance(value, list):
            if not value:
                queryList.append(("", ""))
            else:
                result = self.db._escapeValues(value)
                if not result["OK"]:
                    return result
                query = "( $s )" % ", ".join(result["Value"])
                queryList.append(("IN", query))
        elif isinstance(value, dict):
            for operation, operand in value.items():

                # Prepare the escaped operand first
                if isinstance(operand, list):
                    result = self.db._escapeValues(operand)
                    if not result["OK"]:
                        return result
                    escapedOperand = ", ".join(result["Value"])
                elif isinstance(operand, six.integer_types):
                    escapedOperand = "%d" % operand
                elif isinstance(operand, float):
                    escapedOperand = "%f" % operand
                else:
                    result = self.db._escapeString(operand)
                    if not result["OK"]:
                        return result
                    escapedOperand = result["Value"]

                # Treat the operations
                if operation in [">", "<", ">=", "<="]:
                    if isinstance(operand, list):
                        return S_ERROR("Illegal query: list of values for comparison operation")
                    else:
                        queryList.append((operation, escapedOperand))
                elif operation == "in" or operation == "=":
                    if isinstance(operand, list):
                        queryList.append(("IN", "( %s )" % escapedOperand))
                    else:
                        queryList.append(("=", escapedOperand))
                elif operation == "nin" or operation == "!=":
                    if isinstance(operand, list):
                        queryList.append(("NOT IN", "( %s )" % escapedOperand))
                    else:
                        queryList.append(("!=", escapedOperand))

        return S_OK(queryList)

    def __buildSEQuery(self, storageElements):
        """Return a tuple with table and condition to locate files in a given SE

        :param list storageElements: list of storage element names

        :return: S_OK/S_ERROR, Value - list of tuples (table name, SQL query elements)
        """
        if not storageElements:
            return S_OK([])

        seIDList = []
        for se in storageElements:
            seID = self.db.seNames.get(se, -1)
            if seID == -1:
                return S_ERROR("Unknown SE %s" % se)
            seIDList.append(seID)
        table = "FC_Replicas"
        seString = intListToString(seIDList)
        query = "%%s.SEID IN ( %s )" % seString
        return S_OK([(table, query)])

    def __buildUserMetaQuery(self, userMetaDict):
        """Return a list of tuples with tables and conditions to locate files for a given user Metadata

        :param dict userMetaDict: dictionary with user metadata

        :return: S_OK/S_ERROR, Value - list SQL query elements
        """
        if not userMetaDict:
            return S_OK([])
        resultList = []
        leftJoinTables = []
        for meta, value in userMetaDict.items():
            table = "FC_FileMeta_%s" % meta

            result = self.__createMetaSelection(value)
            if not result["OK"]:
                return result
            for operation, operand in result["Value"]:
                resultList.append((table, "%%s.Value %s %s" % (operation, operand)))
                if operand == "NULL":
                    leftJoinTables.append(table)

        result = S_OK(resultList)
        result["LeftJoinTables"] = leftJoinTables
        return result

    def __buildStandardMetaQuery(self, standardMetaDict):
        """Create SQL query elements for standard file metadata

        :param dict standardMetaDict: dictionary with standard file metadata

        :return: S_OK/S_ERROR, Value - list SQL query elements
        """

        table = "FC_Files"
        queriesFiles = []
        queriesFileInfo = []
        for infield, invalue in standardMetaDict.items():
            value = invalue
            if infield in FILES_TABLE_METAKEYS:
                if infield == "User":
                    value = self.db.users.get(invalue, -1)
                    if value == "-1":
                        return S_ERROR("Unknown user %s" % invalue)
                elif infield == "Group":
                    value = self.db.groups.get(invalue, -1)
                    if value == "-1":
                        return S_ERROR("Unknown group %s" % invalue)

                table = "FC_Files"
                tableIndex = "F"
                field = FILES_TABLE_METAKEYS[infield]
                result = self.__createMetaSelection(value)
                if not result["OK"]:
                    return result
                for operation, operand in result["Value"]:
                    queriesFiles.append("%s.%s %s %s" % (tableIndex, field, operation, operand))
            elif infield in FILEINFO_TABLE_METAKEYS:
                table = "FC_FileInfo"
                tableIndex = "FI"
                field = FILEINFO_TABLE_METAKEYS[infield]
                result = self.__createMetaSelection(value)
                if not result["OK"]:
                    return result
                for operation, operand in result["Value"]:
                    queriesFileInfo.append("%s.%s %s %s" % (tableIndex, field, operation, operand))
            else:
                return S_ERROR("Illegal standard meta key %s" % infield)

        resultList = []
        if queriesFiles:
            query = " AND ".join(queriesFiles)
            resultList.append(("FC_Files", query))
        if queriesFileInfo:
            query = " AND ".join(queriesFileInfo)
            resultList.append(("FC_FileInfo", query))

        return S_OK(resultList)

    def __findFilesByMetadata(self, metaDict, dirList, credDict):
        """Find a list of file IDs meeting the metaDict requirements and belonging
        to directories in dirList

        :param dict metaDict: dictionary with the file metadata
        :param list dirList: list of directories to look into

        :return: S_OK/S_ERROR, Value - list of IDs of found files
        """
        # 1.- classify Metadata keys
        storageElements = None
        standardMetaDict = {}
        userMetaDict = {}
        leftJoinTables = []
        for meta, value in metaDict.items():
            if meta == "SE":
                if isinstance(value, dict):
                    storageElements = value.get("in", [])
                else:
                    storageElements = [value]
            elif meta in FILE_STANDARD_METAKEYS:
                standardMetaDict[meta] = value
            else:
                userMetaDict[meta] = value

        tablesAndConditions = []
        leftJoinTables = []
        # 2.- standard search
        if standardMetaDict:
            result = self.__buildStandardMetaQuery(standardMetaDict)
            if not result["OK"]:
                return result
            tablesAndConditions.extend(result["Value"])
        # 3.- user search
        if userMetaDict:
            result = self.__buildUserMetaQuery(userMetaDict)
            if not result["OK"]:
                return result
            tablesAndConditions.extend(result["Value"])
            leftJoinTables = result["LeftJoinTables"]
        # 4.- SE constraint
        if storageElements:
            result = self.__buildSEQuery(storageElements)
            if not result["OK"]:
                return result
            tablesAndConditions.extend(result["Value"])

        query = "SELECT F.FileID FROM FC_Files F "
        conditions = []
        tables = []

        if dirList:
            dirString = intListToString(dirList)
            conditions.append("F.DirID in (%s)" % dirString)

        counter = 0
        for table, condition in tablesAndConditions:
            if table == "FC_FileInfo":
                query += "INNER JOIN FC_FileInfo FI USING( FileID ) "
                condition = condition.replace("%%", "%")
            elif table == "FC_Files":
                condition = condition.replace("%%", "%")
            else:
                counter += 1
                if table in leftJoinTables:
                    tables.append("LEFT JOIN %s M%d USING( FileID )" % (table, counter))
                else:
                    tables.append("INNER JOIN %s M%d USING( FileID )" % (table, counter))
                table = "M%d" % counter
                condition = condition % table
            conditions.append(condition)

        query += " ".join(tables)
        if conditions:
            query += " WHERE %s" % " AND ".join(conditions)

        result = self.db._query(query)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_OK([])

        #     fileList = [ row[0] for row in result['Value' ] ]
        fileList = []
        for row in result["Value"]:
            fileID = row[0]
            fileList.append(fileID)

        return S_OK(fileList)

    @queryTime
    def findFilesByMetadata(self, metaDict, path, credDict):
        """Find Files satisfying the given metadata

        :param dict metaDict: dictionary with the metaquery parameters
        :param str path: Path to search into
        :param dict credDict: Dictionary with the user credentials

        :return: S_OK/S_ERROR, Value ID:LFN dictionary of selected files
        """
        if not path:
            path = "/"

        # 1.- Get Directories matching the metadata query
        result = self.db.dmeta.findDirIDsByMetadata(metaDict, path, credDict)
        if not result["OK"]:
            return result
        dirList = result["Value"]
        dirFlag = result["Selection"]

        # 2.- Get known file metadata fields
        #     fileMetaDict = {}
        result = self._getFileMetadataFields(credDict)
        if not result["OK"]:
            return result
        fileMetaKeys = list(result["Value"]) + list(FILE_STANDARD_METAKEYS)
        fileMetaDict = dict(item for item in metaDict.items() if item[0] in fileMetaKeys)

        fileList = []
        idLfnDict = {}

        if dirFlag != "None":
            # None means that no Directory satisfies the given query, thus the search is empty
            if dirFlag == "All":
                # All means that there is no Directory level metadata in query, full name space is considered
                dirList = []

            if fileMetaDict:
                # 3.- Do search in File Metadata
                result = self.__findFilesByMetadata(fileMetaDict, dirList, credDict)
                if not result["OK"]:
                    return result
                fileList = result["Value"]
            elif dirList:
                # 4.- if not File Metadata, return the list of files in given directories
                result = self.db.dtree.getFileLFNsInDirectoryByDirectory(dirList, credDict)
                if not result["OK"]:
                    return result
                return S_OK(result["Value"]["IDLFNDict"])
            else:
                # if there is no File Metadata and no Dir Metadata, return an empty list
                idLfnDict = {}

        if fileList:
            # 5.- get the LFN
            result = self.db.fileManager._getFileLFNs(fileList)
            if not result["OK"]:
                return result
            idLfnDict = result["Value"]["Successful"]

        return S_OK(idLfnDict)



#########################################################################################
#
# Script to populate the DIRAC FileCatalog with the information from the LFC
# FileCatalog using multiple LFC sources
#
# Author: A.Tsaregorodtsev
# Last Modified: 9.01.2012
#
#########################################################################################

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

Script.parseCommandLine()

import DIRAC.Resources.Catalog.LcgFileCatalogClient as LcgFileCatalogClient
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getUsernameForDN, getGroupsWithVOMSAttribute
from DIRAC.Core.Utilities.ThreadPool import ThreadPool, ThreadedJob
from DIRAC.Core.Utilities.ProcessPool import ProcessPool
from DIRAC import gConfig, S_OK, S_ERROR

from multiprocessing import Queue, Process, Value, Manager


import time
import sys
import random

dirCount = 0
fileCount = 0
globalStart = time.time()

dnCache = {}
roleCache = {}


def writer(filename, writerQueue, stopFlag):
    print("entering writer")
    with open(filename, "w") as outputFile:
        while not stopFlag.value or not writerQueue.empty():
            outputFile.write(writerQueue.get())

    print("exciting writer stopValue %s" % stopFlag.value)


def getUserNameAndGroup(info):
    """Get the user name and group from the DN and VOMS role"""

    global dnCache, roleCache

    owner = {}
    if "OwnerDN" not in info:
        return owner

    username = dnCache.get(info.get("OwnerDN"))
    if not username:
        result = getUsernameForDN(info.get("OwnerDN", "Unknown"))
        if result["OK"]:
            username = result["Value"]
            dnCache[info["OwnerDN"]] = username
        elif "No username" in result["Message"]:
            username = "Unknown"
            dnCache[info["OwnerDN"]] = username

    if username and username != "Unknown":
        groups = roleCache.get("/" + info.get("OwnerRole"))
        if not groups:
            groups = getGroupsWithVOMSAttribute("/" + info["OwnerRole"])
            roleCache["/" + info["OwnerRole"]] = groups
        if groups:
            owner["username"] = username
            owner["group"] = groups[0]

    return owner


def processDir(initPath, writerQueue, recursive=False, host=None, fcInit=None, dfcInit=None):
    """Process one directory,  possibly recursively"""

    global globalStart, dnCache, roleCache

    # print "AT >>> processDir initPath", initPath

    fc = fcInit
    if not fc:
        fc = LcgFileCatalogClient.LcgFileCatalogClient(host=host)
    dfc = dfcInit
    if not dfc:
        dfc = FileCatalogClient()

    start = time.time()

    resultList = fc.listDirectory(initPath, True)

    if not resultList["OK"]:
        result = S_ERROR("Failed LFC lookup for %s" % initPath)
        result["Path"] = initPath
        return result

    lfc_time = time.time() - start

    s = time.time()

    if resultList["OK"]:
        # Add directories

        if resultList["Value"]["Failed"]:
            result = S_ERROR("Path %s failed: %s" % (initPath, resultList["Value"]["Failed"][initPath]))
            return result

        dirDict = resultList["Value"]["Successful"][initPath]["SubDirs"]
        paths = {}
        for path, info in dirDict.items():
            paths[path] = {}
            paths[path]["Mode"] = info["Mode"]
            owner = getUserNameAndGroup(info)
            if owner:
                paths[path]["Owner"] = owner

        p_dirs = time.time() - s
        s = time.time()
        nDir = len(paths)
        if nDir:
            # print "Adding %d directories in %s" % ( nDir, initPath )
            result = dfc.createDirectory(paths)
            if not result["OK"]:
                print("Error adding directories:%s" % result["Message"])

        e_dirs = time.time() - s

        # Add files

        s = time.time()

        fileDict = resultList["Value"]["Successful"][initPath]["Files"]
        lfns = {}
        for lfn, info in fileDict.items():

            # print info['MetaData']

            lfns[lfn] = {}
            lfns[lfn]["Size"] = info["MetaData"]["Size"]
            lfns[lfn]["Checksum"] = info["MetaData"].get("Checksum", "")
            lfns[lfn]["GUID"] = info["MetaData"]["GUID"]
            lfns[lfn]["Mode"] = info["MetaData"]["Mode"]
            lfns[lfn]["PFN"] = ""
            owner = getUserNameAndGroup(info["MetaData"])
            if owner:
                lfns[lfn]["Owner"] = owner

            if info["Replicas"]:
                seList = list(info["Replicas"])
                lfns[lfn]["SE"] = seList

        p_files = time.time() - s
        s = time.time()

        nFile = len(lfns)
        nRep = 0
        if nFile:

            for lfn in lfns:
                if "SE" in lfns[lfn]:
                    nRep += len(lfns[lfn]["SE"])

            # print "Adding %d files in %s" % ( nFile, initPath )

            done = False
            count = 0
            error = False
            while not done:
                count += 1
                result = dfc.addFile(lfns)
                if not result["OK"]:
                    print("Error adding files %d:" % count, result["Message"])
                    if count > 10:
                        print("Completely failed path %s" % initPath)
                        break
                    error = True
                    time.sleep(2)
                elif error:
                    print("Successfully added files on retry %d" % count)
                    done = True
                else:
                    done = True

        e_files = time.time() - s

        dfc_time = time.time() - start - lfc_time
        total_time = time.time() - globalStart

        format = "== %s: time lfc/dfc %.2f/%.2f, files %d, dirs %d, reps %d, time: %.2f/%.2f/%.2f/%.2f %.2f \n"
        writerQueue.put(
            format % (initPath, lfc_time, dfc_time, nFile, nDir, nRep, p_dirs, e_dirs, p_files, e_files, total_time)
        )
        #     outputFile = open('lfc_dfc.out','a')
        #     outputFile.write( format % (initPath,lfc_time,dfc_time,nFile,nDir,nRep,p_dirs,e_dirs,p_files,e_files,total_time) )
        #     outputFile.close()

        #    print format % (initPath,lfc_time,dfc_time,nFile,fileCount,nDir,dirCount,p_dirs,e_dirs,p_files,e_files,total_time)

        # Go into directories
        if recursive:
            for path in paths:
                result = processDir(path, writerQueue, recursive=True, host=host, fcInit=fc, dfcInit=dfc)
                if result["OK"]:
                    nFile += result["Value"].get("NumberOfFiles", 0)
                    nDir += result["Value"].get("NumberOfDirectories", 0)
                    nRep += result["Value"].get("NumberOfReplicas", 0)

        resultDict = {}
        resultDict["NumberOfFiles"] = nFile
        resultDict["NumberOfDirectories"] = nDir
        resultDict["NumberOfReplicas"] = nRep
        resultDict["Path"] = initPath
        resultDict["Directories"] = list(dirDict)

        # print "AT >>> processDir",initPath,"done %.2f" % (time.time()-start)

        toRet = S_OK(resultDict)
        toRet["writerQueue"] = writerQueue

        return toRet


def finalizeDirectory(task, result):

    global lfcHosts, pPool

    if result["OK"]:
        writerQueue = result["writerQueue"]
        print(
            "Finished directory %(Path)s, dirs: %(NumberOfDirectories)s, files: "
            "%(NumberOfFiles)s, replicas: %(NumberOfReplicas)s" % result["Value"]
        )
        print("%d active tasks remaining" % pPool.getNumWorkingProcesses())

        if "Directories" in result["Value"]:
            for path in result["Value"]["Directories"]:
                random.shuffle(lfcHosts)
                # print pPool.getNumWorkingProcesses(), pPool.hasPendingTasks()
                print("Queueing task for directory %s, lfc %s" % (path, lfcHosts[0]))
                result = pPool.createAndQueueTask(
                    processDir, [path, writerQueue, False, lfcHosts[0]], callback=finalizeDirectory
                )
                if not result["OK"]:
                    print("Failed queueing %s" % path)
    else:
        print("Task failed: %s" % result["Message"])
        if "Path" in result:
            random.shuffle(lfcHosts)
            print("Requeueing task for directory %s, lfc %s" % (result["Path"], lfcHosts[0]))


#########################################################################


pPool = ProcessPool(30, 40, 0)

manager = Manager()
writerQueue = manager.Queue()
stopFlag = Value("i", 0)

# pPool.daemonize()

# lfcHosts = ['lfc-lhcb-ro.cern.ch',
#             'lfc-lhcb-ro.cr.cnaf.infn.it',
#             'lhcb-lfc-fzk.gridka.de',
#             'lfc-lhcb-ro.in2p3.fr',
#             'lfc-lhcb.grid.sara.nl',
#             'lfclhcb.pic.es',
#             'lhcb-lfc.gridpp.rl.ac.uk']
lfcHosts = ["prod-lfc-lhcb-ro.cern.ch"]


# path = "/lhcb/LHCb"
path = "/lhcb/user/c/chaen"
print("Queueing task for directory", path, lfcHosts[0])


writerProc = Process(target=writer, args=("lfc_dfc.out", writerQueue, stopFlag))

writerProc.start()


result = pPool.createAndQueueTask(processDir, [path, writerQueue, False, lfcHosts[0]], callback=finalizeDirectory)
if not result["OK"]:
    print("Failed queueing", path)

for i in range(20):
    pPool.processResults()
    time.sleep(1)

pPool.processAllResults(timeout=300)

stopFlag.value = 1
writerQueue.put("Exit")
writerProc.join()
